"""@package docstring
Iso2Mesh for Python - Mesh data queries and manipulations

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""
__all__ = [
    "brain2mesh",
    "brain1020",
    "intriangulation",
    "label2tpm",
    "tpm2label",
]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
import sys
from iso2mesh.core import v2s, s2m
from iso2mesh.modify import (
    removeisolatednode,
    removedupelem,
    surfboolean,
    meshcheckrepair,
    sms,
    mergemesh,
    slicesurf3,
    slicesurf,
)
from iso2mesh.trait import (
    meshcentroid,
    elemvolume,
    volface,
    layersurf,
    faceneighbors,
    ray2surf,
)
from iso2mesh.geometry import meshabox
from iso2mesh.volume import volgrow, fillholes3d
from iso2mesh.line import polylineinterp, polylinelen, polylinesimplify, closestnode
from iso2mesh.plot import plotmesh

from typing import Tuple, Dict, Union, Optional, List
import matplotlib.pyplot as plt
import warnings
from collections import OrderedDict

##====================================================================================
## implementations
##====================================================================================


def brain2mesh(seg, **cfg):
    """
    Brain2mesh: a one-liner for human brain 3D mesh generation

    Author: Qianqian Fang <q.fang at neu.edu>
    Other contributors: see AUTHORS.txt for details
    Version: 0.8 (Python port)
    URL: http://mcx.space/brain2mesh
    License: GPL version 3
    Reference:
      Anh Phong Tran, Shijie Yan and Qianqian Fang, "Improving model-based
      fNIRS analysis using mesh-based anatomical and light-transport models,"
      Neurophotonics, 7(1), 015008, URL: http://dx.doi.org/10.1117/1.NPh.7.1.015008

    Parameters:
    -----------
    seg : dict or ndarray
        Pre-segmented brain volume (supporting both probalistic tissue
        segmentation and labeled volume). Two formats are accepted:
        1. a dict with subfields (wm,gm,csf,skull,scalp)
           e.g.: seg['wm'], seg['gm'], seg['csf'] represents the white-matter,
           gray-matter and csf segmentations, respectively,
               or
        2. a 4D array for with tissues sorted in outer-to-inner order
           the 4th dimension of the array can 3-6, with the following assumptions
           size(seg,3) == 6 assumes 0-Scalp, 1-Skull, 2-CSF, 3-GM, 4-WM, 5-air pockets
           size(seg,3) == 5 assumes 0-Scalp, 1-Skull, 2-CSF, 3-GM, 4-WM
           size(seg,3) == 4 assumes 0-Scalp, 1-CSF, 2-GM, 3-WM
           size(seg,3) == 3 assumes 0-CSF, 1-GM, 2-WM

    cfg : dict, optional
        Configuration options for the resulting tetrahedral mesh
        cfg['radbound'] : dict with keys {wm,gm,csf,skull,scalp}
           Radius of the Delaunay sphere used in sampling the surfaces.
           Default values are 1.7, 1.7, 2, 2.5 and 3, respectively (reference values for 1x1x1mm^3)
           Scale proportionally for denser volumes. Lower values correspond to denser, higher
           fidelity surface extraction, but also results in denser meshes.
        cfg['maxnode'] : int, default=100000
           Maximum number of nodes extracted for a given surface.
        cfg['maxvol'] : float, default=100
           Maximum volumetric size of elements
        cfg['smooth'] : int, default=0
           Number of iterations to smooth each tissue surface
        cfg['ratio'] : float, default=1.414
           Radius-edge ratio. Lower values increase quality but results in denser meshes
        cfg['dorelabel'] : bool, default=False
           Removes layered meshing assumptions. Currently only works if all five tissue types are present.
        cfg['doairseg'] : bool, default=True
           Within skull layer, additional segmentations can be found.
        cfg['dotruncate'] : str or bool, default='-z' or True
           Truncate mesh in specified direction or disable with 0/False.
        cfg['marginsize'] : int, default=4
           Number of voxels below CSF mesh to truncate when dotruncate is set.
        cfg['imfill'] : str, default='fillholes3d'
           Function name for 3D image hole-filling function

    Returns:
    --------
    brain_n : ndarray
        Node coordinates of the tetrahedral mesh
    brain_el : ndarray
        Element list of the tetrahedral mesh / the last column denotes the boundary ID
    brain_f : ndarray
        Mesh surface element list of the tetrahedral mesh

    Tissue ID for the outputs are as follow:
    0-Air/background, 1-Scalp, 2-Skull, 3-CSF, 4-GM, 5-WM, 6-air pockets

    Reference:
    If you use Brain2Mesh in your publication, the authors of this toolbox
    appreciate if you can cite our Neurophotonics paper listed above.
    """

    # Handling the inputs
    if seg is None:
        print(__doc__)
        return

    # Default density and adaptiveness parameters
    density = {"wm": 2, "gm": 2, "csf": 5, "skull": 4, "scalp": 8}
    adaptiveness = {"wm": 1, "gm": 1, "csf": 1, "skull": 1, "scalp": 1}

    # Parse configuration
    radbound = cfg.get("radbound", density)
    distbound = cfg.get("distbound", adaptiveness)
    qratio = cfg.get("ratio", 1.414)
    maxvol = cfg.get("maxvol", 100)
    maxnode = cfg.get("maxnode", 100000)
    dotruncate = cfg.get("dotruncate", 1)
    dorelabel = cfg.get("dorelabel", 0)
    doairseg = cfg.get("doairseg", 1)
    threshold = cfg.get("threshold", 0.5)
    smooth = cfg.get("smooth", 0)
    surfonly = cfg.get("surfonly", 0)
    marginsize = cfg.get("marginsize", 4)

    segname = list(density.keys())

    # Convert seg to dict format if needed
    if isinstance(seg, dict):
        tpm = seg
    elif isinstance(seg, np.ndarray) and seg.ndim == 4:
        tpm = {}
        for i in range(seg.shape[3]):
            if i < len(segname):
                tpm[segname[i]] = seg[:, :, :, i]
    else:
        raise ValueError("This seg input is currently not supported")

    # Normalizing segmentation inputs to 0-1
    def normalizer(x):
        return np.array(x, dtype=np.float64) / float(np.max(x))

    for key in tpm:
        tpm[key] = normalizer(tpm[key])

    opt = []
    for i in range(len(tpm)):
        opt_i = {"maxnode": maxnode}
        if segname[i] in radbound:
            opt_i["radbound"] = radbound[segname[i]]
        opt_i["distbound"] = distbound[segname[i]]
        opt.append(opt_i)

    cube3 = np.ones((3, 3, 3), dtype=bool)

    # Pre-processing steps to create separations between the tissues in the volume space
    dim = tpm["wm"].shape
    tpm["wm"] = fillholes3d(tpm["wm"] > 0)
    p_wm = tpm["wm"].copy()
    p_pial = p_wm + tpm["gm"]
    p_pial = np.maximum(p_pial, volgrow(p_wm, 1, cube3))
    p_pial = fillholes3d(p_pial > 0)
    expandedGM = p_pial - tpm["wm"] - tpm["gm"]
    expandedGM = volgrow(expandedGM, 1, cube3)

    if "csf" in tpm:
        p_csf = p_pial + tpm["csf"]
        p_csf[p_csf > 1] = 1
        p_csf = np.maximum(p_csf, volgrow(p_pial, 1, cube3))
        expandedCSF = p_csf - tpm["wm"] - tpm["gm"] - tpm["csf"] - expandedGM
        expandedCSF = volgrow(expandedCSF, 1, cube3)

    if "skull" in tpm and "scalp" in tpm and "csf" in tpm:
        p_bone = p_csf + tpm["skull"]
        p_bone[p_bone > 1] = 1
        p_bone = np.maximum(p_bone, volgrow(p_csf, 1, cube3))
        p_skin = p_bone + tpm["scalp"]
        p_skin[p_skin > 1] = 1
        p_skin = np.maximum(p_skin, volgrow(p_bone, 1, cube3))
        expandedSkull = (
            p_bone
            - tpm["wm"]
            - tpm["gm"]
            - tpm["csf"]
            - tpm["skull"]
            - expandedCSF
            - expandedGM
        )
        expandedSkull = volgrow(expandedSkull, 1, cube3)
    elif "scalp" in tpm and "skull" not in tpm:
        p_skin = p_csf + tpm["scalp"]
        p_skin[p_skin > 1] = 1
        p_skin = np.maximum(p_skin, volgrow(p_csf, 1, cube3))
    elif "skull" in tpm and "scalp" not in tpm:
        p_bone = p_csf + tpm["skull"]
        p_bone[p_bone > 1] = 1
        p_bone = np.maximum(p_bone, volgrow(p_csf, 1, cube3))

    # Grayscale/Binary extractions of the surface meshes for the different tissues
    thresh = 0.5
    if not isinstance(threshold, dict):
        thresh = threshold

    wm_thresh = threshold.get("wm", thresh) if isinstance(threshold, dict) else thresh
    gm_thresh = threshold.get("gm", thresh) if isinstance(threshold, dict) else thresh

    wm_n, wm_f, _, _ = v2s(p_wm, wm_thresh, opt[0], "cgalsurf")
    pial_n, pial_f, _, _ = v2s(p_pial, gm_thresh, opt[1], "cgalsurf")
    wm_n, wm_f = meshcheckrepair(wm_n, wm_f[:, :3], "isolated")
    pial_n, pial_f = meshcheckrepair(pial_n, pial_f[:, :3], "isolated")

    if "csf" in tpm:
        csf_thresh = (
            threshold.get("csf", thresh) if isinstance(threshold, dict) else thresh
        )
        csf_n, csf_f, _, _ = v2s(p_csf, csf_thresh, opt[2], "cgalsurf")
        csf_n, csf_f = meshcheckrepair(csf_n, csf_f[:, :3], "isolated")

    if "skull" in tpm:
        optskull = {"radbound": radbound["skull"], "maxnode": maxnode}
        skull_thresh = (
            threshold.get("skull", thresh) if isinstance(threshold, dict) else thresh
        )
        bone_n, bone_f, _, _ = v2s(p_bone, skull_thresh, optskull, "cgalsurf")

        bone_node, el_bone, _ = s2m(
            bone_n, bone_f, 1.0, maxvol, "tetgen1.5", None, None, "-A"
        )
        unique_labels = np.unique(el_bone[:, 4])  # 5th column is index 4 (0-based)
        vol_bone = []
        for i, label in enumerate(unique_labels):
            vol_bone.append(
                np.sum(elemvolume(bone_node, el_bone[el_bone[:, 4] == label, :4]))
            )

        maxval = np.max(vol_bone)
        I = np.argmax(vol_bone)
        max_label = unique_labels[I]

        if len(unique_labels) > 1:
            no_air2 = bone_node
            el_air2 = el_bone[el_bone[:, 4] != max_label, :]
            no_air2, el_air2, _ = removeisolatednode(no_air2, el_air2)
            f_air2 = volface(el_air2[:, :4])[0]

        bone_n2 = bone_node
        bone_f2 = volface(el_bone[:, :4])[0]
        bone_f2 = removedupelem(bone_f2)
        bone_n2, bone_f2, _ = removeisolatednode(bone_n2, bone_f2)
        if doairseg == 0:
            bone_n = bone_n2
            bone_f = bone_f2

    if "scalp" in tpm:
        optscalp = {"radbound": radbound["scalp"], "maxnode": maxnode}
        scalp_thresh = (
            threshold.get("scalp", thresh) if isinstance(threshold, dict) else thresh
        )
        skin_n, skin_f, _, _ = v2s(p_skin, scalp_thresh, optscalp, "cgalsurf")

    # Smoothing step
    if isinstance(smooth, dict) or smooth > 0:
        scount = 0
        if not isinstance(smooth, dict):
            scount = smooth

        wm_smooth = smooth.get("wm", scount) if isinstance(smooth, dict) else scount
        if wm_smooth > 0:
            wm_n = sms(wm_n, wm_f[:, :3], wm_smooth, 0.5, "lowpass")
            wm_n, wm_f = meshcheckrepair(wm_n, wm_f[:, :3], "meshfix")

        gm_smooth = smooth.get("gm", scount) if isinstance(smooth, dict) else scount
        if gm_smooth > 0:
            pial_n = sms(pial_n, pial_f[:, :3], gm_smooth, 0.5, "lowpass")
            pial_n, pial_f = meshcheckrepair(pial_n, pial_f[:, :3], "meshfix")

        if "csf" in tpm:
            csf_smooth = (
                smooth.get("csf", scount) if isinstance(smooth, dict) else scount
            )
            if csf_smooth > 0:
                csf_n = sms(csf_n, csf_f[:, :3], csf_smooth, 0.5, "lowpass")
                csf_n, csf_f = meshcheckrepair(csf_n, csf_f[:, :3], "meshfix")

        if "skull" in tpm:
            skull_smooth = (
                smooth.get("skull", scount) if isinstance(smooth, dict) else scount
            )
            if skull_smooth > 0:
                bone_n = sms(bone_n, bone_f[:, :3], skull_smooth, 0.5, "lowpass")
                bone_n, bone_f = meshcheckrepair(bone_n, bone_f[:, :3], "meshfix")

        if "scalp" in tpm:
            scalp_smooth = (
                smooth.get("scalp", scount) if isinstance(smooth, dict) else scount
            )
            if scalp_smooth > 0:
                skin_n = sms(skin_n, skin_f[:, :3], scalp_smooth, 0.5, "lowpass")
                skin_n, skin_f = meshcheckrepair(skin_n, skin_f[:, :3], "meshfix")

    # Surface only mode
    if surfonly == 1:
        wm_f = np.column_stack([wm_f, np.ones(wm_f.shape[0])])
        pial_f = np.column_stack([pial_f, np.full(pial_f.shape[0], 2)])
        brain_n, brain_el = mergemesh(wm_n, wm_f, pial_n, pial_f)

        if "csf" in tpm:
            csf_f = np.column_stack([csf_f, np.full(csf_f.shape[0], 3)])
            brain_n, brain_el = mergemesh(brain_n, brain_el, csf_n, csf_f)
            if "skull" in tpm:
                bone_f = np.column_stack([bone_f, np.full(bone_f.shape[0], 4)])
                brain_n, brain_el = mergemesh(brain_n, brain_el, bone_n, bone_f)
                if "scalp" in tpm:
                    skin_f = np.column_stack([skin_f, np.full(skin_f.shape[0], 5)])
                    brain_n, brain_el = mergemesh(brain_n, brain_el, skin_n, skin_f)

        labels, ia, ib = np.unique(
            brain_el[:, 3], return_index=True, return_inverse=True
        )
        labels = np.arange(5, 5 - len(labels), -1)
        brain_el[:, 3] = labels[ib]
        brain_f = np.array([])
        return brain_n, brain_el, brain_f

    # Main loop for the meshing pipeline
    for loop in range(2):
        # If the first pass fails, a second pass is called using the decoupled function
        # to eliminate intersections between surface meshes
        if (loop == 1) and ("label_elem" in locals()):
            continue
        if (loop == 1) and ("label_elem" not in locals()):
            if "bone_n" in locals() and "skin_n" in locals():
                bone_n, bone_f = surfboolean(
                    bone_n[:, :3],
                    bone_f[:, :3],
                    "decouple",
                    skin_n[:, :3],
                    skin_f[:, :3],
                )
            if "bone_n" in locals() and "csf_n" in locals():
                csf_n, csf_f = surfboolean(
                    csf_n[:, :3], csf_f[:, :3], "decouple", bone_n[:, :3], bone_f[:, :3]
                )
            if "pial_n" in locals() and "csf_n" in locals():
                pial_n, pial_f = surfboolean(
                    pial_n[:, :3], pial_f[:, :3], "decouple", csf_n[:, :3], csf_f[:, :3]
                )
            if "pial_n" in locals() and "wm_n" in locals():
                wm_n, wm_f = surfboolean(
                    wm_n[:, :3], wm_f[:, :3], "decouple", pial_n[:, :3], pial_f[:, :3]
                )

        if "wm" in tpm and "gm" in tpm:
            surf_n, surf_f = surfboolean(
                wm_n[:, :3], wm_f[:, :3], "resolve", pial_n, pial_f
            )
        if "csf" in tpm:
            surf_n, surf_f = surfboolean(surf_n, surf_f, "resolve", csf_n, csf_f)
        if "skull" in tpm:
            surf_n, surf_f = surfboolean(surf_n, surf_f, "resolve", bone_n, bone_f)
        if "scalp" in tpm:
            surf_n, surf_f = surfboolean(surf_n, surf_f, "resolve", skin_n, skin_f)

        final_surf_n = surf_n
        final_surf_f = surf_f

        if surfonly == 2:
            brain_n = final_surf_n
            brain_el = final_surf_f
            brain_f = np.array([])
            return brain_n, brain_el, brain_f

        # If the whole head option is deactivated, the cut is made at the base of the brain using a box cutting
        if dotruncate == 1 or isinstance(dotruncate, str):
            dim_max = np.max(surf_n, axis=0)
            if "csf" in tpm:
                dim2 = np.min(csf_n, axis=0)
            else:
                dim2 = np.min(surf_n, axis=0)

            if dotruncate == 1 or dotruncate == "-z":
                nbox, fbox, ebox = meshabox(
                    [-1, -1, dim2[2] + marginsize],
                    [dim_max[0] + 1, dim_max[1] + 1, dim_max[2] + 1],
                    500,
                )
            elif dotruncate == "-y":
                nbox, fbox, ebox = meshabox(
                    [-1, dim2[1] + marginsize, -1],
                    [dim_max[0] + 1, dim_max[1] + 1, dim_max[2] + 1],
                    500,
                )
            elif dotruncate == "-x":
                nbox, fbox, ebox = meshabox(
                    [dim2[0] + marginsize, -1, -1],
                    [dim_max[0] + 1, dim_max[1] + 1, dim_max[2] + 1],
                    500,
                )
            elif dotruncate == "+z":
                nbox, fbox, ebox = meshabox(
                    [-1, -1, -1],
                    [dim_max[0] + 1, dim_max[1] + 1, dim2[2] - marginsize],
                    500,
                )
            elif dotruncate == "+y":
                nbox, fbox, ebox = meshabox(
                    [-1, -1, -1],
                    [dim_max[0] + 1, dim2[1] - marginsize, dim_max[2] + 1],
                    500,
                )
            elif dotruncate == "+x":
                nbox, fbox, ebox = meshabox(
                    [-1, -1, -1],
                    [dim2[0] - marginsize, dim_max[1] + 1, dim_max[2] + 1],
                    500,
                )

            fbox = volface(ebox)[0]
            nbox, fbox, _ = removeisolatednode(nbox, fbox)
            final_surf_n, final_surf_f = surfboolean(
                nbox, fbox, "first", surf_n, surf_f
            )

        if surfonly == 3:
            brain_n = final_surf_n
            brain_el = final_surf_f
            brain_f = np.array([])
            return brain_n, brain_el, brain_f

        # Generates a coarse tetrahedral mesh of the combined tissues
        try:
            final_n, final_e, _ = s2m(
                final_surf_n, final_surf_f, 1.0, maxvol, "tetgen1.5", None, None, "-A"
            )
        except:
            print(
                "volumetric mesh generation failed, returning the intermediate surface model only"
            )
            brain_n = final_surf_n
            brain_f = final_surf_f
            brain_el = np.array([])
            return brain_n, brain_el, brain_f

        # Removes the elements that are part of the box, but not the brain/head
        if dotruncate == 1 or isinstance(dotruncate, str):
            maxval_idx = np.argmax(final_n, axis=0)
            k = np.where(final_e[:, :4] == maxval_idx[2] + 1)[0]
            if len(k) > 0:
                k = k[0]
                exclude_label = final_e[k % len(final_e), 4]
                final_e = final_e[final_e[:, 4] != exclude_label, :]
                final_n, final_e, _ = removeisolatednode(final_n, final_e)

        # Here the labels created through the coarse mesh generated through Tetgen are saved
        # with the centroid of one of the elements for intriangulation testing later
        label, label_elem = np.unique(final_e[:, 4], return_index=True)
        label_centroid = meshcentroid(final_n, final_e[label_elem, :4])

        if "scalp" in tpm:
            no_skin, el_skin, _ = s2m(
                skin_n, skin_f, 1.0, maxvol, "tetgen1.5", None, None, "-A"
            )
            unique_skin_labels = np.unique(el_skin[:, 4])
            vol_skin = []
            for i, skin_label in enumerate(unique_skin_labels):
                vol_skin.append(
                    np.sum(
                        elemvolume(no_skin, el_skin[el_skin[:, 4] == skin_label, :4])
                    )
                )

            maxval = np.max(vol_skin)
            I = np.argmax(vol_skin)
            max_skin_label = unique_skin_labels[I]

            if len(unique_skin_labels) > 1:
                no_air = no_skin
                el_air = el_skin[el_skin[:, 4] != max_skin_label, :]
                no_air, el_air, _ = removeisolatednode(no_air, el_air)
                f_air = volface(el_air[:, :4])[0]
                f_air = removedupelem(f_air)

            el_skin = el_skin[el_skin[:, 4] == max_skin_label, :]
            no_skin, el_skin, _ = removeisolatednode(no_skin, el_skin)
            f_skin = volface(el_skin[:, :4])[0]
            f_skin = removedupelem(f_skin)

        # When the label_elem does not exist, it often indicates a failure at the generation of a coarse
        # tetrahedral mesh. The alternative meshing pathway using decoupling is then called to make a
        # second attempt at creating the combined tetrahedral mesh.
        if "label_elem" not in locals() and loop == 0:
            print(
                "Initial meshing procedure failed. The option parameter might need to be adjusted."
            )
            print("Activating alternative meshing pathway...")
            continue

        # The labels are given to each of the tissues
        # WM(1) - GM(2) - CSF(3) - Bone(4) - Scalp(5) - Air(6)
        newlabel = np.zeros(len(label_elem))
        if "bone_n" in locals() and "no_air2" in locals():
            # Need to subtract 1 when using as indices since faces/elements are 1-indexed
            newlabel = intriangulation(no_air2, f_air2[:, :3], label_centroid)
        if "no_skin" in locals() and "no_air" in locals():
            newlabel = newlabel | intriangulation(no_air, f_air[:, :3], label_centroid)

        newlabel = newlabel.astype(float)
        idx = np.where(newlabel == 0)[0]

        newtag = np.zeros(len(idx))
        newtag = intriangulation(wm_n, wm_f[:, :3], label_centroid[idx, :]) * 6
        newtag = np.maximum(
            newtag,
            intriangulation(pial_n, pial_f[:, :3], label_centroid[idx, :]) * 5,
        )
        if "csf_n" in locals():
            newtag = np.maximum(
                newtag,
                intriangulation(csf_n, csf_f[:, :3], label_centroid[idx, :]) * 4,
            )
        if "bone_n2" in locals():
            newtag = np.maximum(
                newtag,
                intriangulation(bone_n2, bone_f2[:, :3], label_centroid[idx, :]) * 3,
            )
        if "no_skin" in locals():
            newtag = np.maximum(
                newtag,
                intriangulation(no_skin, f_skin[:, :3], label_centroid[idx, :]) * 2,
            )

        newlabel[idx] = newtag
        newlabel = 7 - newlabel
        final_e[:, 4] = newlabel[final_e[:, 4] - 1]  # Subtract 1 for 0-based indexing

        # This step consolidates adjacent labels of the same tissue
        new_label = np.unique(final_e[:, 4])
        face = np.empty((0, 3), dtype=int)
        for i in new_label:
            face = np.vstack([face, volface(final_e[final_e[:, 4] == i, :4])[0]])
        face = np.sort(face, axis=1)
        face = np.unique(face, axis=0)
        node, face, _ = removeisolatednode(final_n, face)

        # The final mesh is generated here with the desired properties
        cmdopt = f"-A -pq{qratio}a{maxvol}"
        brain_n, brain_el, _ = s2m(
            node, face, 1.0, maxvol, "tetgen1.5", None, None, cmdopt
        )

        label2, label_brain_el = np.unique(brain_el[:, 4], return_index=True)
        label_centroid2 = meshcentroid(brain_n, brain_el[label_brain_el, :4])

        # The labeling process is repeated for the final mesh
        # WM(1) - GM(2) - CSF(3) - Bone(4) - Scalp(5) - Air(6)
        newlabel = np.zeros(len(label_brain_el))
        if "bone_n" in locals() and "no_air2" in locals():
            newlabel = intriangulation(no_air2, f_air2[:, :3], label_centroid2)
        if "no_skin" in locals() and "no_air" in locals():
            newlabel = newlabel | intriangulation(no_air, f_air[:, :3], label_centroid2)

        newlabel = newlabel.astype(float)
        idx = np.where(newlabel == 0)[0]

        newtag = np.zeros(len(idx))
        newtag = intriangulation(wm_n, wm_f[:, :3], label_centroid2[idx, :]) * 6
        newtag = np.maximum(
            newtag,
            intriangulation(pial_n, pial_f[:, :3], label_centroid2[idx, :]) * 5,
        )
        if "csf_n" in locals():
            newtag = np.maximum(
                newtag,
                intriangulation(csf_n, csf_f[:, :3], label_centroid2[idx, :]) * 4,
            )
        if "bone_n2" in locals():
            newtag = np.maximum(
                newtag,
                intriangulation(bone_n2, bone_f2[:, :3], label_centroid2[idx, :]) * 3,
            )
        if "no_skin" in locals():
            newtag = np.maximum(
                newtag,
                intriangulation(no_skin, f_skin[:, :3], label_centroid2[idx, :]) * 2,
            )

        newlabel[idx] = newtag
        newlabel = 7 - newlabel
        brain_el[:, 4] = newlabel[brain_el[:, 4] - 1]  # Subtract 1 for 0-based indexing

        break  # Exit the loop after successful completion

    # Relabeling step to remove layered assumptions
    if dorelabel == 1 and ("skull" in tpm and "scalp" in tpm):
        centroid = meshcentroid(brain_n[:, :3], brain_el[:, :4])
        centroid = np.ceil(centroid).astype(int)
        tag = np.zeros(len(brain_el))
        facenb = faceneighbors(brain_el[:, :4])

        for i in range(len(brain_el)):
            cx, cy, cz = centroid[i]
            if (expandedGM[cx, cy, cz] > 0.5) and (brain_el[i, 4] == 2):
                if tpm["scalp"][cx, cy, cz] > 0.5:
                    brain_el[i, 4] = 5
                elif tpm["skull"][cx, cy, cz] > 0.5:
                    brain_el[i, 4] = 4
                else:
                    brain_el[i, 4] = 3
                tag[i] = 1
                for j in range(4):
                    if facenb[i, j] > 0:
                        tag[facenb[i, j] - 1] = 1  # Subtract 1 for 0-based indexing
            elif (expandedCSF[cx, cy, cz] > 0.5) and (brain_el[i, 4] == 3):
                if tpm["scalp"][cx, cy, cz] > 0.5:
                    brain_el[i, 4] = 5
                else:
                    brain_el[i, 4] = 4
                tag[i] = 1
                for j in range(4):
                    if facenb[i, j] > 0:
                        tag[facenb[i, j] - 1] = 1  # Subtract 1 for 0-based indexing
            elif (expandedSkull[cx, cy, cz] > 0.5) and (brain_el[i, 4] == 4):
                brain_el[i, 4] = 5
                tag[i] = 1
                for j in range(4):
                    if facenb[i, j] > 0:
                        tag[facenb[i, j] - 1] = 1  # Subtract 1 for 0-based indexing

        labels = np.zeros((len(brain_el), 4))
        labels2 = np.zeros((len(brain_el), 6))
        for i in range(len(brain_el)):
            for j in range(4):
                if facenb[i, j] > 0:
                    neighbor_idx = facenb[i, j] - 1  # Subtract 1 for 0-based indexing
                    labels[i, j] = brain_el[neighbor_idx, 4]
                    labels2[i, int(labels[i, j])] += 1
                else:
                    labels[i, j] = 0

        max_labels = np.argmax(labels2, axis=1)
        max_counts = np.max(labels2, axis=1)

        for i in range(len(brain_el)):
            if tag[i] == 1:
                if (max_counts[i] > 2) and (brain_el[i, 4] != max_labels[i]):
                    brain_el[i, 4] = max_labels[i]

    brain_el[:, 4] = 6 - brain_el[:, 4]

    # Generate brain_f if requested
    brain_f = np.array([])
    if cfg.get("face", False):  # Check if brain_f is requested
        brain_f = layersurf(brain_el)[0]

    return brain_n, brain_el, brain_f


def intriangulation(
    vertices: np.ndarray, faces: np.ndarray, testp: np.ndarray, heavytest: int = 0
) -> np.ndarray:
    """
    Test points in 3d whether inside or outside a (closed) triangulation
    Usage: in = intriangulation(vertices, faces, testp, heavytest)

    Arguments (input):
    vertices   - points in 3d as matrix with three columns
    faces      - description of triangles as matrix with three columns.
                Each row contains three indices into the matrix of vertices
                which gives the three cornerpoints of the triangle.
    testp      - points in 3d as matrix with three columns
    heavytest  - int n >= 0. Perform n additional randomized rotation tests.

    IMPORTANT: the set of vertices and faces has to form a watertight surface!

    Arguments (output):
    in - a vector of length size(testp,0), containing 0 and 1.
         in[nr] =  0: testp[nr,:] is outside the triangulation
         in[nr] =  1: testp[nr,:] is inside the triangulation
         in[nr] = -1: unable to decide for testp[nr,:]

    Thanks to Adam A for providing the FEX submission voxelise. The
    algorithms of voxelise form the algorithmic kernel of intriangulation.

    Thanks to Sven to discussions about speed and avoiding problems in
    special cases.

    Author: Johannes Korsawe, heavily based on voxelise from Adam A.
    E-mail: johannes.korsawe@volkswagen.de
    Python conversion: Preserves exact algorithm from MATLAB version
    Release: 1.3
    Release date: 25/09/2013
    """

    # Check number of inputs
    if vertices is None or faces is None or testp is None:
        print("??? Error using ==> intriangulation\nThree input matrices are needed.\n")
        return np.array([])

    if heavytest is None:
        heavytest = 0

    # Check size of inputs
    if vertices.shape[1] != 3 or faces.shape[1] != 3 or testp.shape[1] != 3:
        print(
            "??? Error using ==> intriangulation\nAll input matrices must have three columns.\n"
        )
        return np.array([])

    # Convert faces from 1-based to 0-based indexing when used as indices
    # Note: we keep faces array unchanged, but subtract 1 when indexing
    ipmax = np.max(faces)
    zerofound = np.any(faces == 0)
    if ipmax > vertices.shape[0] or zerofound:
        print(
            "??? Error using ==> intriangulation\nThe triangulation data is defect. use trisurf(faces,vertices[:,0],vertices[:,1],vertices[:,2]) for test of deficiency.\n"
        )
        return np.array([])

    # Loop for heavytest
    inreturn = np.zeros((testp.shape[0], 1))
    VER = vertices.copy()
    TESTP = testp.copy()

    for n in range(1, heavytest + 2):  # MATLAB: for n = 1:heavytest + 1

        # Randomize
        if n > 1:
            v = np.random.rand(1, 3)
            D = rotmatrix(v / np.linalg.norm(v), np.random.rand() * 180 / np.pi)
            vertices = VER @ D
            testp = TESTP @ D
        else:
            vertices = VER.copy()

        # Preprocessing data
        meshXYZ = np.zeros((faces.shape[0], 3, 3))
        for loop in range(3):
            # Subtract 1 when using faces as indices (convert from 1-based to 0-based)
            meshXYZ[:, :, loop] = vertices[faces[:, loop] - 1, :]

        # Basic idea (ingenious from FeX-submission voxelise):
        # If point is inside, it will cross the triangulation an uneven number of times in each direction (x, -x, y, -y, z, -z).

        # The function VOXELISEinternal is about 98% identical to its version inside voxelise.m.
        # This includes the elaborate comments. Thanks to Adam A!

        # z-direction:
        # initialization of results and correction list
        in_result, cl = VOXELISEinternal(testp[:, 0], testp[:, 1], testp[:, 2], meshXYZ)

        # x-direction:
        # has only to be done for those points, that were not determinable in the first step --> cl
        if len(cl) > 0:
            in2, cl2 = VOXELISEinternal(
                testp[cl, 1], testp[cl, 2], testp[cl, 0], meshXYZ[:, [1, 2, 0], :]
            )
            # Use results of x-direction that determined "inside"
            in_result[cl[in2 == 1]] = 1
            # remaining indices with unclear result
            cl = cl[cl2]

        # y-direction:
        # has only to be done for those points, that were not determinable in the first and second step --> cl
        if len(cl) > 0:
            in3, cl3 = VOXELISEinternal(
                testp[cl, 2], testp[cl, 0], testp[cl, 1], meshXYZ[:, [2, 0, 1], :]
            )

            # Use results of y-direction that determined "inside"
            in_result[cl[in3 == 1]] = 1
            # remaining indices with unclear result
            cl = cl[cl3]

        # Mark those indices, where all three tests have failed
        in_result[cl] = -1

        if n == 1:
            inreturn = in_result.copy()  # Starting guess
        else:
            # if ALWAYS inside, use as inside!
            # I = find(inreturn ~= in);
            # inreturn(I(in(I)==0)) = 0;

            # if AT LEAST ONCE inside, use as inside!
            I = np.where(inreturn.flatten() != in_result.flatten())[0]
            inreturn.flatten()[I[in_result.flatten()[I] == 1]] = 1

    return inreturn.flatten().astype(int)


def VOXELISEinternal(
    testx: np.ndarray, testy: np.ndarray, testz: np.ndarray, meshXYZ: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Internal voxelization function - exact port from MATLAB voxelise

    This function is about 98% identical to its version inside voxelise.m.
    This includes the elaborate comments. Thanks to Adam A!
    """

    # Prepare logical array to hold the logical data:
    OUTPUT = np.zeros((testx.shape[0], 1))

    # Identify the min and max x,y coordinates of the mesh:
    meshZmin = np.min(meshXYZ[:, 2, :])
    meshZmax = np.max(meshXYZ[:, 2, :])

    # Identify the min and max x,y,z coordinates of each facet:
    meshXYZmin = np.min(meshXYZ, axis=2)
    meshXYZmax = np.max(meshXYZ, axis=2)

    # ======================================================
    # TURN OFF DIVIDE-BY-ZERO WARNINGS
    # ======================================================
    # This prevents the Y1predicted, Y2predicted, Y3predicted and YRpredicted
    # calculations creating divide-by-zero warnings. Suppressing these warnings
    # doesn't affect the code, because only the sign of the result is important.
    # That is, 'Inf' and '-Inf' results are ok.
    # The warning will be returned to its original state at the end of the code.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)

        # ======================================================
        # START COMPUTATION
        # ======================================================

        correctionLIST = (
            []
        )  # Prepare to record all rays that fail the voxelisation. This array is built on-the-fly, but since
        # it ought to be relatively small should not incur too much of a speed penalty.

        # Loop through each testpoint.
        # The testpoint-array will be tested by passing rays in the z-direction through
        # each x,y coordinate of the testpoints, and finding the locations where the rays cross the mesh.
        facetCROSSLIST = np.zeros(1000, dtype=int)  # uses countindex: nf
        nm = meshXYZmin.shape[0]

        for loop in range(len(OUTPUT)):

            nf = 0
            # - 1a - Find which mesh facets could possibly be crossed by the ray:
            # possibleCROSSLISTy = find( meshXYZmin(:,2)<=testy(loop) & meshXYZmax(:,2)>=testy(loop) );

            # - 1b - Find which mesh facets could possibly be crossed by the ray:
            # possibleCROSSLIST = possibleCROSSLISTy( meshXYZmin(possibleCROSSLISTy,1)<=testx(loop) & meshXYZmax(possibleCROSSLISTy,1)>=testx(loop) );

            # Do - 1a - and - 1b - faster
            possibleCROSSLISTy = np.where(
                (testy[loop] - meshXYZmin[:, 1]) * (meshXYZmax[:, 1] - testy[loop]) > 0
            )[0]
            possibleCROSSLISTx = (testx[loop] - meshXYZmin[possibleCROSSLISTy, 0]) * (
                meshXYZmax[possibleCROSSLISTy, 0] - testx[loop]
            ) > 0
            possibleCROSSLIST = possibleCROSSLISTy[possibleCROSSLISTx]

            if (
                len(possibleCROSSLIST) != 0
            ):  # Only continue the analysis if some nearby facets were actually identified

                # - 2 - For each facet, check if the ray really does cross the facet rather than just passing it close-by:

                # GENERAL METHOD:
                # 1. Take each edge of the facet in turn.
                # 2. Find the position of the opposing vertex to that edge.
                # 3. Find the position of the ray relative to that edge.
                # 4. Check if ray is on the same side of the edge as the opposing vertex.
                # 5. If this is true for all three edges, then the ray definitely passes through the facet.
                #
                # NOTES:
                # 1. If the ray crosses exactly on an edge, this is counted as crossing the facet.
                # 2. If a ray crosses exactly on a vertex, this is also taken into account.

                for loopCHECKFACET in possibleCROSSLIST:

                    # Check if ray crosses the facet. This method is much (>>10 times) faster than using the built-in function 'inpolygon'.
                    # Taking each edge of the facet in turn, check if the ray is on the same side as the opposing vertex. If so, let testVn=1

                    Y1predicted = meshXYZ[loopCHECKFACET, 1, 1] - (
                        (meshXYZ[loopCHECKFACET, 1, 1] - meshXYZ[loopCHECKFACET, 1, 2])
                        * (
                            meshXYZ[loopCHECKFACET, 0, 1]
                            - meshXYZ[loopCHECKFACET, 0, 0]
                        )
                        / (
                            meshXYZ[loopCHECKFACET, 0, 1]
                            - meshXYZ[loopCHECKFACET, 0, 2]
                        )
                    )
                    YRpredicted = meshXYZ[loopCHECKFACET, 1, 1] - (
                        (meshXYZ[loopCHECKFACET, 1, 1] - meshXYZ[loopCHECKFACET, 1, 2])
                        * (meshXYZ[loopCHECKFACET, 0, 1] - testx[loop])
                        / (
                            meshXYZ[loopCHECKFACET, 0, 1]
                            - meshXYZ[loopCHECKFACET, 0, 2]
                        )
                    )

                    if (
                        (
                            Y1predicted > meshXYZ[loopCHECKFACET, 1, 0]
                            and YRpredicted > testy[loop]
                        )
                        or (
                            Y1predicted < meshXYZ[loopCHECKFACET, 1, 0]
                            and YRpredicted < testy[loop]
                        )
                        or (
                            meshXYZ[loopCHECKFACET, 1, 1]
                            - meshXYZ[loopCHECKFACET, 1, 2]
                        )
                        * (meshXYZ[loopCHECKFACET, 0, 1] - testx[loop])
                        == 0
                    ):
                        # testV1 = 1;   # The ray is on the same side of the 2-3 edge as the 1st vertex.
                        pass
                    else:
                        # testV1 = 0;   # The ray is on the opposite side of the 2-3 edge to the 1st vertex.
                        # As the check is for ALL three checks to be true, we can continue here, if only one check fails
                        continue

                    Y2predicted = meshXYZ[loopCHECKFACET, 1, 2] - (
                        (meshXYZ[loopCHECKFACET, 1, 2] - meshXYZ[loopCHECKFACET, 1, 0])
                        * (
                            meshXYZ[loopCHECKFACET, 0, 2]
                            - meshXYZ[loopCHECKFACET, 0, 1]
                        )
                        / (
                            meshXYZ[loopCHECKFACET, 0, 2]
                            - meshXYZ[loopCHECKFACET, 0, 0]
                        )
                    )
                    YRpredicted = meshXYZ[loopCHECKFACET, 1, 2] - (
                        (meshXYZ[loopCHECKFACET, 1, 2] - meshXYZ[loopCHECKFACET, 1, 0])
                        * (meshXYZ[loopCHECKFACET, 0, 2] - testx[loop])
                        / (
                            meshXYZ[loopCHECKFACET, 0, 2]
                            - meshXYZ[loopCHECKFACET, 0, 0]
                        )
                    )

                    if (
                        (
                            Y2predicted > meshXYZ[loopCHECKFACET, 1, 1]
                            and YRpredicted > testy[loop]
                        )
                        or (
                            Y2predicted < meshXYZ[loopCHECKFACET, 1, 1]
                            and YRpredicted < testy[loop]
                        )
                        or (
                            meshXYZ[loopCHECKFACET, 1, 2]
                            - meshXYZ[loopCHECKFACET, 1, 0]
                        )
                        * (meshXYZ[loopCHECKFACET, 0, 2] - testx[loop])
                        == 0
                    ):
                        # testV2 = 1;   # The ray is on the same side of the 3-1 edge as the 2nd vertex.
                        pass
                    else:
                        # testV2 = 0;   # The ray is on the opposite side of the 3-1 edge to the 2nd vertex.
                        # As the check is for ALL three checks to be true, we can continue here, if only one check fails
                        continue

                    Y3predicted = meshXYZ[loopCHECKFACET, 1, 0] - (
                        (meshXYZ[loopCHECKFACET, 1, 0] - meshXYZ[loopCHECKFACET, 1, 1])
                        * (
                            meshXYZ[loopCHECKFACET, 0, 0]
                            - meshXYZ[loopCHECKFACET, 0, 2]
                        )
                        / (
                            meshXYZ[loopCHECKFACET, 0, 0]
                            - meshXYZ[loopCHECKFACET, 0, 1]
                        )
                    )
                    YRpredicted = meshXYZ[loopCHECKFACET, 1, 0] - (
                        (meshXYZ[loopCHECKFACET, 1, 0] - meshXYZ[loopCHECKFACET, 1, 1])
                        * (meshXYZ[loopCHECKFACET, 0, 0] - testx[loop])
                        / (
                            meshXYZ[loopCHECKFACET, 0, 0]
                            - meshXYZ[loopCHECKFACET, 0, 1]
                        )
                    )

                    if (
                        (
                            Y3predicted > meshXYZ[loopCHECKFACET, 1, 2]
                            and YRpredicted > testy[loop]
                        )
                        or (
                            Y3predicted < meshXYZ[loopCHECKFACET, 1, 2]
                            and YRpredicted < testy[loop]
                        )
                        or (
                            meshXYZ[loopCHECKFACET, 1, 0]
                            - meshXYZ[loopCHECKFACET, 1, 1]
                        )
                        * (meshXYZ[loopCHECKFACET, 0, 0] - testx[loop])
                        == 0
                    ):
                        # testV3 = 1;   # The ray is on the same side of the 1-2 edge as the 3rd vertex.
                        pass
                    else:
                        # testV3 = 0;   # The ray is on the opposite side of the 1-2 edge to the 3rd vertex.
                        # As the check is for ALL three checks to be true, we can continue here, if only one check fails
                        continue

                    nf = nf + 1
                    if nf <= len(facetCROSSLIST):
                        facetCROSSLIST[nf - 1] = loopCHECKFACET
                    else:
                        # Expand array if needed
                        facetCROSSLIST = np.concatenate(
                            [facetCROSSLIST, np.zeros(1000, dtype=int)]
                        )
                        facetCROSSLIST[nf - 1] = loopCHECKFACET

                # Use only values ~=0
                facetCROSSLIST = facetCROSSLIST[:nf]

                # - 3 - Find the z coordinate of the locations where the ray crosses each facet:
                gridCOzCROSS = np.zeros(nf)
                for i, loopFINDZ in enumerate(facetCROSSLIST):

                    # METHOD:
                    # 1. Define the equation describing the plane of the facet. For a
                    # more detailed outline of the maths, see:
                    # http://local.wasp.uwa.edu.au/~pbourke/geometry/planeeq/
                    #    Ax + By + Cz + D = 0
                    #    where  A = y1 (z2 - z3) + y2 (z3 - z1) + y3 (z1 - z2)
                    #           B = z1 (x2 - x3) + z2 (x3 - x1) + z3 (x1 - x2)
                    #           C = x1 (y2 - y3) + x2 (y3 - y1) + x3 (y1 - y2)
                    #           D = - x1 (y2 z3 - y3 z2) - x2 (y3 z1 - y1 z3) - x3 (y1 z2 - y2 z1)
                    # 2. For the x and y coordinates of the ray, solve these equations to find the z coordinate in this plane.

                    planecoA = (
                        meshXYZ[loopFINDZ, 1, 0]
                        * (meshXYZ[loopFINDZ, 2, 1] - meshXYZ[loopFINDZ, 2, 2])
                        + meshXYZ[loopFINDZ, 1, 1]
                        * (meshXYZ[loopFINDZ, 2, 2] - meshXYZ[loopFINDZ, 2, 0])
                        + meshXYZ[loopFINDZ, 1, 2]
                        * (meshXYZ[loopFINDZ, 2, 0] - meshXYZ[loopFINDZ, 2, 1])
                    )

                    planecoB = (
                        meshXYZ[loopFINDZ, 2, 0]
                        * (meshXYZ[loopFINDZ, 0, 1] - meshXYZ[loopFINDZ, 0, 2])
                        + meshXYZ[loopFINDZ, 2, 1]
                        * (meshXYZ[loopFINDZ, 0, 2] - meshXYZ[loopFINDZ, 0, 0])
                        + meshXYZ[loopFINDZ, 2, 2]
                        * (meshXYZ[loopFINDZ, 0, 0] - meshXYZ[loopFINDZ, 0, 1])
                    )

                    planecoC = (
                        meshXYZ[loopFINDZ, 0, 0]
                        * (meshXYZ[loopFINDZ, 1, 1] - meshXYZ[loopFINDZ, 1, 2])
                        + meshXYZ[loopFINDZ, 0, 1]
                        * (meshXYZ[loopFINDZ, 1, 2] - meshXYZ[loopFINDZ, 1, 0])
                        + meshXYZ[loopFINDZ, 0, 2]
                        * (meshXYZ[loopFINDZ, 1, 0] - meshXYZ[loopFINDZ, 1, 1])
                    )

                    planecoD = (
                        -meshXYZ[loopFINDZ, 0, 0]
                        * (
                            meshXYZ[loopFINDZ, 1, 1] * meshXYZ[loopFINDZ, 2, 2]
                            - meshXYZ[loopFINDZ, 1, 2] * meshXYZ[loopFINDZ, 2, 1]
                        )
                        - meshXYZ[loopFINDZ, 0, 1]
                        * (
                            meshXYZ[loopFINDZ, 1, 2] * meshXYZ[loopFINDZ, 2, 0]
                            - meshXYZ[loopFINDZ, 1, 0] * meshXYZ[loopFINDZ, 2, 2]
                        )
                        - meshXYZ[loopFINDZ, 0, 2]
                        * (
                            meshXYZ[loopFINDZ, 1, 0] * meshXYZ[loopFINDZ, 2, 1]
                            - meshXYZ[loopFINDZ, 1, 1] * meshXYZ[loopFINDZ, 2, 0]
                        )
                    )

                    if abs(planecoC) < 1e-14:
                        planecoC = 0

                    gridCOzCROSS[i] = (
                        -planecoD - planecoA * testx[loop] - planecoB * testy[loop]
                    ) / planecoC

                if len(gridCOzCROSS) == 0:
                    continue

                # Remove values of gridCOzCROSS which are outside of the mesh limits (including a 1e-12 margin for error).
                gridCOzCROSS = gridCOzCROSS[
                    (gridCOzCROSS >= meshZmin - 1e-12)
                    & (gridCOzCROSS <= meshZmax + 1e-12)
                ]

                # Round gridCOzCROSS to remove any rounding errors, and take only the unique values:
                gridCOzCROSS = np.round(gridCOzCROSS * 1e10) / 1e10

                # Replacement of the call to unique (gridCOzCROSS = unique(gridCOzCROSS);) by the following line:
                tmp = np.sort(gridCOzCROSS)
                if len(tmp) > 1:
                    I = np.concatenate(([True], tmp[1:] - tmp[:-1] != 0))
                    gridCOzCROSS = tmp[I]
                else:
                    gridCOzCROSS = tmp

                # - 4 - Label as being inside the mesh all the voxels that the ray passes through after crossing one facet before crossing another facet:

                if (
                    len(gridCOzCROSS) % 2 == 0
                ):  # Only rays which cross an even number of facets are voxelised

                    for loopASSIGN in range(1, len(gridCOzCROSS) // 2 + 1):
                        voxelsINSIDE = (
                            testz[loop] > gridCOzCROSS[2 * loopASSIGN - 2]
                            and testz[loop] < gridCOzCROSS[2 * loopASSIGN - 1]
                        )
                        OUTPUT[loop] = int(voxelsINSIDE)
                        if voxelsINSIDE:
                            break

                elif (
                    len(gridCOzCROSS) != 0
                ):  # Remaining rays which meet the mesh in some way are not voxelised, but are labelled for correction later.
                    correctionLIST.append(loop)

    # ======================================================
    # RESTORE DIVIDE-BY-ZERO WARNINGS TO THE ORIGINAL STATE
    # ======================================================
    # (handled automatically by context manager)

    # J.Korsawe: A correction is not possible as the testpoints need not to be
    #            ordered in any way.
    #            voxelise contains a correction algorithm which is appended here
    #            without changes in syntax.

    return OUTPUT.flatten().astype(int), np.array(correctionLIST, dtype=int)


def rotmatrix(v: np.ndarray, deg: float) -> np.ndarray:
    """
    Calculate the rotation matrix about v by deg degrees
    """

    deg = deg / 180 * np.pi
    if deg != 0:
        v = v / np.linalg.norm(v)
        v1 = v[0, 0] if v.ndim > 1 else v[0]
        v2 = v[0, 1] if v.ndim > 1 else v[1]
        v3 = v[0, 2] if v.ndim > 1 else v[2]
        ca = np.cos(deg)
        sa = np.sin(deg)
        D = np.array(
            [
                [
                    ca + v1 * v1 * (1 - ca),
                    v1 * v2 * (1 - ca) - v3 * sa,
                    v1 * v3 * (1 - ca) + v2 * sa,
                ],
                [
                    v2 * v1 * (1 - ca) + v3 * sa,
                    ca + v2 * v2 * (1 - ca),
                    v2 * v3 * (1 - ca) - v1 * sa,
                ],
                [
                    v3 * v1 * (1 - ca) - v2 * sa,
                    v3 * v2 * (1 - ca) + v1 * sa,
                    ca + v3 * v3 * (1 - ca),
                ],
            ]
        )
    else:
        D = np.eye(3)

    return D


def brain1020(
    node: np.ndarray,
    face: np.ndarray,
    initpoints: Optional[Union[Dict, np.ndarray]] = None,
    perc1: int = 10,
    perc2: int = 20,
    **kwargs,
) -> Tuple[Dict[str, np.ndarray], Dict[str, np.ndarray], np.ndarray]:
    """
    Compute 10-20-like scalp landmarks with user-specified density on a head mesh

    Author: Qianqian Fang (q.fang at neu.edu)
    Python conversion: Preserves exact algorithm from MATLAB version

    Parameters:
    -----------
    node : ndarray
        Full head mesh node list
    face : ndarray
        Full head mesh element list- a 3-column array defines face list
        for the exterior (scalp) surface; a 4-column array defines the
        tetrahedral mesh of the full head.
    initpoints : dict or ndarray, optional
        One can provide the 3-D coordinates of the below
        5 landmarks: nz, iz, lpa, rpa, cz0 (cz0 is the initial guess of cz)
        initpoints can be a dict with the above landmark names
        as keys, or a 5x3 array defining these points in the above
        mentioned order (one can use the output landmarks as initpoints)
    perc1 : int, optional
        The percentage of geodesic distance towards the rim of
        the landmarks; this is the first number of the 10-20 or 10-10 or
        10-5 systems, in this case, it is 10 (for 10%). Default is 10.
    perc2 : int, optional
        The percentage of geodesic distance towards the center
        of the landmarks; this is the 2nd number of the 10-20 or 10-10 or
        10-5 systems, which are 20, 10, 5, respectively, default is 20
    **kwargs : dict
        Additional options:
        'display' : bool, default=True
            If True, plot landmarks and curves
        'cztol' : float, default=1e-6
            The tolerance for searching cz that bisects saggital and coronal reference curves
        'maxcziter' : int, default=10
            The maximum number of iterations to update cz to bisect both cm and sm curves
        'baseplane' : bool, default=True
            If True, create the reference curves along the primary control points (nz,iz,lpa,rpa)
        'minangle' : float, default=0
            If set to a positive number, this specifies the minimum angle (radian) between
            adjacent segments in the reference curves to avoid sharp turns

    Returns:
    --------
    landmarks : dict
        A dictionary storing all computed landmarks. The keys include two sections:
        1) 'nz','iz','lpa','rpa','cz': individual 3D positions defining
           the 5 principle reference points: nasion (nz), inion (iz),
           left-pre-auricular-point (lpa), right-pre-auricular-point
           (rpa) and vertex (cz) - cz is updated from initpoints to bisect
           the saggital and coronal ref. curves.
        2) landmarks along specific cross-sections, each cross section
           may contain more than 1 position.
    curves : dict
        A dictionary storing all computed cross-section curves. The
        keys are named similarly to landmarks, except that
        landmarks stores the 10-? points, and curves stores the
        detailed cross-sectional curves
    initpoints : ndarray
        A 5x3 array storing the principle reference points in the
        orders of 'nz','iz','lpa','rpa','cz'

    Notes:
    ------
    This function is part of brain2mesh toolbox (http://mcx.space/brain2mesh)
    License: GPL v3 or later, see LICENSE.txt for details

    This function requires a pre-installed Iso2Mesh Toolbox
    Download URL: http://github.com/fangq/iso2mesh
    Website: http://iso2mesh.sf.net
    """

    if node is None or face is None:
        raise ValueError("one must provide a head-mesh to call this function")

    if node.size == 0 or face.size == 0 or face.shape[1] <= 2 or node.shape[1] < 3:
        raise ValueError(
            "input node must have 3 columns, face must have at least 3 columns"
        )

    # Parse user options
    showplot = kwargs.get("display", False)
    baseplane = kwargs.get("baseplane", True)
    tol = kwargs.get("cztol", 1e-6)
    dosimplify = kwargs.get("minangle", 0)
    maxcziter = kwargs.get("maxcziter", 10)

    if isinstance(initpoints, list):
        initpoints = np.array(initpoints)

    # Handle case where initpoints has only 3 landmarks (nz, lpa, rpa)
    if initpoints is not None and (
        (isinstance(initpoints, dict) and "iz" not in initpoints)
        or (isinstance(initpoints, np.ndarray) and initpoints.shape[0] == 3)
    ):

        if isinstance(initpoints, dict):
            nz = np.array(initpoints["nz"]).flatten()
            lpa = np.array(initpoints["lpa"]).flatten()
            rpa = np.array(initpoints["rpa"]).flatten()
        else:
            nz = initpoints[0, :]
            lpa = initpoints[1, :]
            rpa = initpoints[2, :]

        # This assumes nz, lpa, rpa, iz are on the same plane to find iz on the head surface
        pa_mid = np.mean([lpa, rpa], axis=0)
        v0 = pa_mid - nz
        iz, e0 = ray2surf(node, face, nz, v0, ">")

        # To find cz, we assume that the vector from iz nz midpoint to cz is perpendicular
        # to the plane defined by nz, lpa, and rpa.
        iznz_mid = (nz + iz) * 0.5
        v0 = np.cross(nz - rpa, lpa - rpa)
        cz, e0 = ray2surf(node, face, iznz_mid, v0, ">")

        if isinstance(initpoints, dict):
            initpoints["iz"] = iz
            initpoints["cz"] = cz
        else:
            initpoints = np.vstack(
                [
                    initpoints[0:1, :],
                    iz.reshape(1, -1),
                    initpoints[1:3, :],
                    cz.reshape(1, -1),
                ]
            )

    # Convert initpoints input to a 5x3 array
    if isinstance(initpoints, dict):
        landmarks = {
            "nz": np.array(initpoints["nz"]).flatten(),
            "iz": np.array(initpoints["iz"]).flatten(),
            "lpa": np.array(initpoints["lpa"]).flatten(),
            "rpa": np.array(initpoints["rpa"]).flatten(),
            "cz": np.array(initpoints["cz"]).flatten(),
        }
        initpoints = np.vstack(
            [
                landmarks["nz"],
                landmarks["iz"],
                landmarks["lpa"],
                landmarks["rpa"],
                landmarks["cz"],
            ]
        )

    # Convert tetrahedral mesh into a surface mesh
    if face.shape[1] >= 4:
        face = volface(face[:, :4])[0]  # Use first 4 columns for tetrahedral faces

    # Remove nodes not located in the surface
    node, face, _ = removeisolatednode(node, face)

    # If initpoints is not sufficient, we need interactive selection
    # For now, we'll require full initpoints or raise an error
    if initpoints is None or initpoints.shape[0] < 5:
        raise ValueError(
            "initpoints must contain 5 landmarks: nz, iz, lpa, rpa, cz. "
            + "Interactive selection not implemented in Python version."
        )

    if showplot:
        print("Initial points:")
        print(initpoints)

    # Save input initpoints to landmarks output, cz is not finalized
    if initpoints.shape[0] >= 5:
        landmarks = {
            "nz": initpoints[0, :],
            "iz": initpoints[1, :],
            "lpa": initpoints[2, :],
            "rpa": initpoints[3, :],
            "cz": initpoints[4, :],
        }

    # At this point, initpoints contains {nz, iz, lpa, rpa, cz0}
    # Plot the head mesh
    if showplot:
        hh = plotmesh(node, face, alpha=0, color="wheat", linewidth=0.1)
        ax = hh["ax"][0]
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")
        ax.set_title("Brain 10-20 System Landmarks")

    lastcz = np.array([1, 1, 1]) * np.inf
    cziter = 0
    curves = {}

    # Find cz that bisects cm and sm curves within a tolerance, using UI 10-10 approach
    while np.linalg.norm(initpoints[4, :] - lastcz) > tol and cziter < maxcziter:

        # Step 1: nz, iz and cz0 to determine saggital reference curve
        nsagg = slicesurf(node, face, initpoints[[0, 1, 4], :])

        # Step 1.1: get cz1 as the mid-point between iz and nz
        slen, nsagg, _ = polylinelen(
            nsagg, initpoints[0, :], initpoints[1, :], initpoints[4, :]
        )
        if dosimplify:
            nsagg, slen = polylinesimplify(nsagg, dosimplify)
        idx, weight, cz = polylineinterp(slen, np.sum(slen) * 0.5, nsagg)
        initpoints[4, :] = cz[0, :]

        # Step 1.2: lpa, rpa and cz1 to determine coronal reference curve, update cz1
        curves["cm"] = slicesurf(node, face, initpoints[[2, 3, 4], :])
        len_cm, curves["cm"], _ = polylinelen(
            curves["cm"], initpoints[2, :], initpoints[3, :], initpoints[4, :]
        )
        if dosimplify:
            curves["cm"], len_cm = polylinesimplify(curves["cm"], dosimplify)
        idx, weight, coro = polylineinterp(len_cm, np.sum(len_cm) * 0.5, curves["cm"])
        lastcz = initpoints[4, :].copy()
        initpoints[4, :] = coro[0, :]
        cziter += 1
        if showplot:
            print(
                f"cz iteration {cziter} error {np.linalg.norm(initpoints[4, :] - lastcz):.2e}"
            )

    # Set the finalized cz to output
    landmarks["cz"] = initpoints[4, :].copy()

    if showplot:
        print("Finalized points:")
        print(initpoints)

    # Step 2: subdivide saggital (sm) and coronal (cm) ref curves
    perc_range = np.arange(perc1, 100 - perc1 + perc2, perc2) * 0.01
    idx, weight, coro = polylineinterp(
        len_cm, np.sum(len_cm) * perc_range, curves["cm"]
    )
    landmarks["cm"] = coro  # t7, c3, cz, c4, t8

    curves["sm"] = slicesurf(node, face, initpoints[[0, 1, 4], :])
    slen, curves["sm"], _ = polylinelen(
        curves["sm"], initpoints[0, :], initpoints[1, :], initpoints[4, :]
    )
    if dosimplify:
        curves["sm"], slen = polylinesimplify(curves["sm"], dosimplify)
    idx, weight, sagg = polylineinterp(slen, np.sum(slen) * perc_range, curves["sm"])
    landmarks["sm"] = sagg  # fpz, fz, cz, pz, oz

    # Step 3: fpz, t7 and oz to determine left 10% axial reference curve
    landmarks["aal"], curves["aal"], landmarks["apl"], curves["apl"] = slicesurf3(
        node,
        face,
        landmarks["sm"][0, :],
        landmarks["cm"][0, :],
        landmarks["sm"][-1, :],
        perc2 * 2,
    )

    # Step 4: fpz, t8 and oz to determine right 10% axial reference curve
    landmarks["aar"], curves["aar"], landmarks["apr"], curves["apr"] = slicesurf3(
        node,
        face,
        landmarks["sm"][0, :],
        landmarks["cm"][-1, :],
        landmarks["sm"][-1, :],
        perc2 * 2,
    )

    # Show plots of the landmarks
    if showplot:
        ax.plot(
            curves["sm"][:, 0],
            curves["sm"][:, 1],
            curves["sm"][:, 2],
            "r-",
            linewidth=2,
            label="Sagittal",
        )
        ax.plot(
            curves["cm"][:, 0],
            curves["cm"][:, 1],
            curves["cm"][:, 2],
            "g-",
            linewidth=2,
            label="Coronal",
        )
        ax.plot(
            curves["aal"][:, 0],
            curves["aal"][:, 1],
            curves["aal"][:, 2],
            "k-",
            linewidth=1,
        )
        ax.plot(
            curves["aar"][:, 0],
            curves["aar"][:, 1],
            curves["aar"][:, 2],
            "k-",
            linewidth=1,
        )
        ax.plot(
            curves["apl"][:, 0],
            curves["apl"][:, 1],
            curves["apl"][:, 2],
            "b-",
            linewidth=1,
        )
        ax.plot(
            curves["apr"][:, 0],
            curves["apr"][:, 1],
            curves["apr"][:, 2],
            "b-",
            linewidth=1,
        )

        ax.scatter(
            landmarks["sm"][:, 0],
            landmarks["sm"][:, 1],
            landmarks["sm"][:, 2],
            c="red",
            s=50,
            marker="o",
            label="Sagittal landmarks",
        )
        ax.scatter(
            landmarks["cm"][:, 0],
            landmarks["cm"][:, 1],
            landmarks["cm"][:, 2],
            c="green",
            s=50,
            marker="o",
            label="Coronal landmarks",
        )
        ax.scatter(
            landmarks["aal"][:, 0],
            landmarks["aal"][:, 1],
            landmarks["aal"][:, 2],
            c="black",
            s=50,
            marker="o",
        )
        ax.scatter(
            landmarks["aar"][:, 0],
            landmarks["aar"][:, 1],
            landmarks["aar"][:, 2],
            c="magenta",
            s=50,
            marker="o",
        )
        ax.scatter(
            landmarks["apl"][:, 0],
            landmarks["apl"][:, 1],
            landmarks["apl"][:, 2],
            c="black",
            s=50,
            marker="o",
        )
        ax.scatter(
            landmarks["apr"][:, 0],
            landmarks["apr"][:, 1],
            landmarks["apr"][:, 2],
            c="magenta",
            s=50,
            marker="o",
        )

    # Step 5: computing all anterior coronal cuts, moving away from the medial cut (cm) toward frontal
    idxcz = closestnode(landmarks["sm"], landmarks["cz"])[0]
    skipcount = int(np.floor(10 / perc2))

    for i in range(
        1, landmarks["aal"].shape[0] - skipcount + 1
    ):  # MATLAB: for i = 1:size(landmarks.aal, 1) - skipcount
        step = (
            (perc2 * 25)
            * 0.1
            * (
                1
                + (
                    (perc2 < 20 + perc2 < 10)
                    and i == landmarks["aal"].shape[0] - skipcount
                )
            )
        )
        # Subtract 1 from idxcz when using as index (convert from 1-based to 0-based)
        cal_landmarks, leftpart, car_landmarks, rightpart = slicesurf3(
            node,
            face,
            landmarks["aal"][i - 1, :],
            landmarks["sm"][idxcz - 1 - i, :],
            landmarks["aar"][i - 1, :],
            step,
        )

        landmarks[f"cal_{i}"] = cal_landmarks
        landmarks[f"car_{i}"] = car_landmarks

        if showplot:
            ax.plot(leftpart[:, 0], leftpart[:, 1], leftpart[:, 2], "k-", linewidth=1)
            ax.plot(
                rightpart[:, 0], rightpart[:, 1], rightpart[:, 2], "k-", linewidth=1
            )
            ax.scatter(
                cal_landmarks[:, 0],
                cal_landmarks[:, 1],
                cal_landmarks[:, 2],
                c="yellow",
                s=30,
                marker="o",
            )
            ax.scatter(
                car_landmarks[:, 0],
                car_landmarks[:, 1],
                car_landmarks[:, 2],
                c="cyan",
                s=30,
                marker="o",
            )

    # Step 6: computing all posterior coronal cuts, moving away from the medial cut (cm) toward occipital
    for i in range(
        1, landmarks["apl"].shape[0] - skipcount + 1
    ):  # MATLAB: for i = 1:size(landmarks.apl, 1) - skipcount
        step = (
            (perc2 * 25)
            * 0.1
            * (
                1
                + (
                    (perc2 < 20 + perc2 < 10)
                    and i == landmarks["apl"].shape[0] - skipcount
                )
            )
        )
        # Subtract 1 from idxcz when using as index (convert from 1-based to 0-based)
        cpl_landmarks, leftpart, cpr_landmarks, rightpart = slicesurf3(
            node,
            face,
            landmarks["apl"][i - 1, :],
            landmarks["sm"][idxcz - 1 + i, :],
            landmarks["apr"][i - 1, :],
            step,
        )

        landmarks[f"cpl_{i}"] = cpl_landmarks
        landmarks[f"cpr_{i}"] = cpr_landmarks

        if showplot:
            ax.plot(leftpart[:, 0], leftpart[:, 1], leftpart[:, 2], "k-", linewidth=1)
            ax.plot(
                rightpart[:, 0], rightpart[:, 1], rightpart[:, 2], "k-", linewidth=1
            )
            ax.scatter(
                cpl_landmarks[:, 0],
                cpl_landmarks[:, 1],
                cpl_landmarks[:, 2],
                c="yellow",
                s=30,
                marker="o",
            )
            ax.scatter(
                cpr_landmarks[:, 0],
                cpr_landmarks[:, 1],
                cpr_landmarks[:, 2],
                c="cyan",
                s=30,
                marker="o",
            )

    # Step 7: create the axial cuts across principle ref. points: left: nz, lpa, iz, right: nz, rpa, iz
    if baseplane and perc2 <= 10:
        (
            landmarks["paal"],
            curves["paal"],
            landmarks["papl"],
            curves["papl"],
        ) = slicesurf3(
            node, face, landmarks["nz"], landmarks["lpa"], landmarks["iz"], perc2 * 2
        )
        (
            landmarks["paar"],
            curves["paar"],
            landmarks["papr"],
            curves["papr"],
        ) = slicesurf3(
            node, face, landmarks["nz"], landmarks["rpa"], landmarks["iz"], perc2 * 2
        )

        if showplot:
            ax.plot(
                curves["paal"][:, 0],
                curves["paal"][:, 1],
                curves["paal"][:, 2],
                "k-",
                linewidth=1,
            )
            ax.plot(
                curves["paar"][:, 0],
                curves["paar"][:, 1],
                curves["paar"][:, 2],
                "k-",
                linewidth=1,
            )
            ax.plot(
                curves["papl"][:, 0],
                curves["papl"][:, 1],
                curves["papl"][:, 2],
                "k-",
                linewidth=1,
            )
            ax.plot(
                curves["papr"][:, 0],
                curves["papr"][:, 1],
                curves["papr"][:, 2],
                "k-",
                linewidth=1,
            )

            ax.scatter(
                landmarks["paal"][:, 0],
                landmarks["paal"][:, 1],
                landmarks["paal"][:, 2],
                c="yellow",
                s=30,
                marker="o",
            )
            ax.scatter(
                landmarks["papl"][:, 0],
                landmarks["papl"][:, 1],
                landmarks["papl"][:, 2],
                c="cyan",
                s=30,
                marker="o",
            )
            ax.scatter(
                landmarks["paar"][:, 0],
                landmarks["paar"][:, 1],
                landmarks["paar"][:, 2],
                c="yellow",
                s=30,
                marker="o",
            )
            ax.scatter(
                landmarks["papr"][:, 0],
                landmarks["papr"][:, 1],
                landmarks["papr"][:, 2],
                c="cyan",
                s=30,
                marker="o",
            )

    return landmarks, curves, initpoints


def label2tpm(
    vol: np.ndarray, names: Optional[Union[List[str], Dict[int, str]]] = None
) -> Dict[str, np.ndarray]:
    """
    Converting a multi-label volume to binary tissue probabilistic maps (TPMs)

    Author: Qianqian Fang (q.fang at neu.edu)
    Python conversion: Preserves exact algorithm from MATLAB version

    Parameters:
    -----------
    vol : ndarray
        A 2-D or 3-D array of integer values.
    names : list or dict, optional
        A list of strings defining the names of each label,
        alternatively, a dictionary with label (integer) as
        the key and name as the value. The names can be a subset of all
        labels.

    Returns:
    --------
    seg : dict
        A dictionary with subfields of 3D or 4D uint8 array; the subfield
        names are defined via the optional names input; the unnamed
        labels will be named as 'label_#'.

    Notes:
    ------
    This function is part of brain2mesh toolbox (http://mcx.space/brain2mesh)
    License: GPL v3 or later, see LICENSE.txt for details
    """

    if not np.issubdtype(vol.dtype, np.number):
        raise TypeError("input must be a numerical array")

    # Get unique values excluding 0, equivalent to setdiff(sort(unique(vol(:))), 0)
    vol_flat = vol.flatten(order="F")  # Flatten using Fortran order as requested
    unique_vals = np.unique(vol_flat)
    val = np.setdiff1d(unique_vals, 0)  # Remove 0 from unique values
    val = np.sort(val)  # Ensure sorted order

    # Check if input should be converted to labels first
    if len(val) > vol.size * (0.1**vol.ndim):
        raise ValueError("please convert the input to labels first")

    seg = {}
    for i in range(
        len(val)
    ):  # MATLAB: for i = 1:length(val), Python: for i in range(len(val))
        nm = f"label_{val[i]}"  # sprintf('label_%d', val(i)) equivalent

        if names is not None:
            if isinstance(names, list):
                # MATLAB: if (i <= length(names)), Python: if (i < len(names))
                if i < len(names):
                    nm = names[i]
            elif isinstance(names, dict):
                # MATLAB: elseif (isa(names, 'containers.Map') && isKey(names, val(i)))
                if val[i] in names:
                    nm = names[val[i]]

        # MATLAB: seg.(nm) = uint8(vol == val(i))
        seg[nm] = (vol == val[i]).astype(np.uint8)

    return seg


def tpm2label(
    seg: Union[Dict[str, np.ndarray], List[np.ndarray], np.ndarray],
    segorder: Optional[List[str]] = None,
) -> Union[np.ndarray, Tuple[np.ndarray, List[str]]]:
    """
    Converting tissue probabilistic maps (TPMs) to a multi-label volume

    Author: Qianqian Fang (q.fang at neu.edu)
    Python conversion: Preserves exact algorithm from MATLAB version

    Parameters:
    -----------
    seg : dict, list, or ndarray
        A dictionary, a list, or a 3D or 4D array; if seg is a list
        or a dictionary, their elements must be 2D/3D arrays of the same
        sizes;
    segorder : list, optional
        If seg is a dictionary, segorder allows one to assign output
        labels using customized order instead of the creation order

    Returns:
    --------
    vol : ndarray
        A 2-D or 3-D array of the same type/size of the input arrays. The
        label for each voxel is determined by the index to the highest
        value in TPM of the same voxel. If a voxel is a background voxel
        - i.e. zeros for all TPMs, it stays 0
    names : list, optional
        A list storing the names of the labels (if input is a
        dictionary), the first string is the name for label 1, and so on
        (only returned if input is a dictionary)

    Notes:
    ------
    This function is part of brain2mesh toolbox (http://mcx.space/brain2mesh)
    License: GPL v3 or later, see LICENSE.txt for details
    """

    mask = seg
    names = []

    # Handle dictionary input (equivalent to isstruct(seg) in MATLAB)
    if isinstance(seg, dict):
        if segorder is not None:
            # Reorder fields according to segorder (equivalent to orderfields(seg, segorder))
            seg = orderfields_python(seg, segorder)

        # Get field names (equivalent to fieldnames(seg))
        names = list(seg.keys())

        # Extract values using cellfun equivalent
        # MATLAB: mask = cellfun(@(x) seg.(x), names, 'UniformOutput', false);
        mask = [seg[x] for x in names]

    # Handle list input (equivalent to iscell(mask) in MATLAB)
    if isinstance(mask, list):
        # Concatenate along new dimension (equivalent to cat(ndims(mask{1}) + 1, mask{:}))
        if len(mask) > 0:
            # Get number of dimensions of first element
            ndims_first = mask[0].ndim
            # Concatenate along the next dimension
            mask = np.stack(mask, axis=ndims_first)
        else:
            mask = np.array([])

    # Check if input is numeric (equivalent to isnumeric(mask) in MATLAB)
    if not np.issubdtype(mask.dtype, np.number):
        raise TypeError(
            "input must be a list/dict array with numeric elements of matching dimensions"
        )

    # Find maximum values and indices along last dimension
    # MATLAB: [newmask, vol] = max(mask, [], ndims(mask));
    last_dim = (
        mask.ndim - 1
    )  # ndims(mask) in MATLAB is mask.ndim in Python, but for axis we need ndim-1
    newmask = np.max(mask, axis=last_dim)
    vol = (
        np.argmax(mask, axis=last_dim) + 1
    )  # +1 to convert from 0-based to 1-based indexing

    # Set background voxels to 0 (equivalent to vol .* (sum(mask, ndims(mask)) > 0))
    background_mask = np.sum(mask, axis=last_dim) > 0
    vol = vol * background_mask.astype(vol.dtype)

    # Return results based on input type
    if isinstance(seg, dict):
        return vol, names
    else:
        return vol


def orderfields_python(struct_dict, order: List[str]):
    """
    Python equivalent of MATLAB's orderfields function

    Reorders dictionary fields according to specified order.

    Parameters:
    -----------
    struct_dict : dict
        Dictionary to reorder
    order : list
        List specifying the desired field order

    Returns:
    --------
    ordered_dict : dict
        Dictionary with fields reordered according to order
    """

    ordered_dict = OrderedDict()

    # First add fields in the specified order
    for field_name in order:
        if field_name in struct_dict:
            ordered_dict[field_name] = struct_dict[field_name]

    # Then add any remaining fields that weren't in the order list
    for field_name, field_value in struct_dict.items():
        if field_name not in ordered_dict:
            ordered_dict[field_name] = field_value

    return ordered_dict



"""@package docstring
Iso2Mesh for Python - Primitive shape meshing functions

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""

__all__ = [
    "plotsurf",
    "plotasurf",
    "plotmesh",
    "plotedges",
    "plottetra",
    "plotvolume",
]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
import matplotlib.pyplot as plt
from iso2mesh.trait import volface, meshcentroid

COLOR_OFFSET = 3
# _________________________________________________________________________________________________________


def plotsurf(node, face, *args, **kwargs):
    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    from matplotlib.colors import Normalize

    rngstate = np.random.get_state()

    randseed = int("623F9A9E", 16) + COLOR_OFFSET

    if "ISO2MESH_RANDSEED" in globals():
        randseed = globals()["ISO2MESH_RANDSEED"]
    np.random.seed(randseed)

    sc = np.random.rand(10, 3)

    ax = plt.gca()

    h = {"fig": [], "ax": [], "obj": []}
    h["fig"].append(plt.gcf())
    h["ax"].append(ax)

    if not "color" in kwargs and not "cmap" in kwargs:
        kwargs["cmap"] = plt.get_cmap("jet")

    if isinstance(face, list):  # polyhedral facets
        newsurf = {}
        colormap = []

        for fc in face:
            if (
                isinstance(fc, (list, tuple))
                and len(fc) >= 2
                and isinstance(fc[0], (list, tuple))
            ):
                group_id = fc[1][0]
                if group_id + 1 > sc.shape[0]:
                    sc = np.vstack([sc, np.random.rand(group_id + 1 - sc.shape[0], 3)])
                newsurf.setdefault(group_id, []).append(np.asarray(fc[0]) - 1)
            else:
                newsurf.setdefault(1, []).append(np.asarray(fc) - 1)

        polydata = [
            node[np.array(subf).flatten(), :3]
            for subface in newsurf.values()
            for subf in subface
        ]
        if node.shape[1] > 3:
            node_values = node[:, 3]
            face_values = np.array(
                [
                    np.mean(node_values[np.array(subf).flatten()])
                    for subface in newsurf.values()
                    for subf in subface
                ]
            )
            norm = Normalize(vmin=face_values.min(), vmax=face_values.max())
            colormap = kwargs["cmap"](norm(face_values))
        else:
            colormap = [
                sc[i - 1, :] for i, subface in newsurf.items() for subf in subface
            ]

    elif face.shape[1] == 2:
        h = plotedges(node, face, *args, **kwargs)
        return h
    elif face.shape[1] == 4:
        tag = face[:, 3]
        types = np.unique(tag)

        if len(types) > sc.shape[0]:
            sc = np.vstack([sc, np.random.rand(len(types) - sc.shape[0], 3)])

        # plt.hold(True)
        polydata = []
        colormap = []
        for i in range(len(types)):
            pdata, _ = plotasurf(
                node,
                face[tag == types[i], 0:3],
                *args,
                **kwargs,
            )
            polydata.extend(pdata)
            colormap.extend([sc[i].tolist()] * len(pdata))
    else:
        polydata, colormap = plotasurf(node, face, *args, **kwargs)

    if "colormap" in locals() and len(colormap) > 0 and not "facecolors" in kwargs:
        kwargs["facecolors"] = colormap

    if "cmap" in kwargs and not "facecolors" in kwargs and face:
        node_values = node[:, 3] if node.shape[1] > 3 else node[:, 2]
        face_values = np.array([np.mean(node_values[f]) for f in face[:, :3] - 1])
        norm = Normalize(vmin=face_values.min(), vmax=face_values.max())
        kwargs["facecolors"] = kwargs["cmap"](norm(face_values))

    if not "linewidth" in kwargs:
        kwargs["linewidth"] = 0.3

    patch = Poly3DCollection(polydata, edgecolors="k", **kwargs)

    ax.add_collection3d(patch)
    _autoscale_3d(ax, node)
    h["obj"].append(patch)

    np.random.set_state(rngstate)
    # plt.axis("equal")

    return h


# _________________________________________________________________________________________________________


def plotasurf(node, face, *args, **kwargs):
    from matplotlib.colors import Normalize

    poly3d = [[node[i, :3] for i in p] for p in face[:, :3] - 1]
    node_values = node[:, 3] if node.shape[1] > 3 else node[:, 2]
    face_values = np.array([np.mean(node_values[f]) for f in face[:, :3] - 1])
    norm = Normalize(vmin=face_values.min(), vmax=face_values.max())
    colmap = []
    if "cmap" in kwargs:
        colmap = kwargs["cmap"](norm(face_values))
    return poly3d, colmap


# _________________________________________________________________________________________________________


def plottetra(node, elem, *args, **kwargs):
    """
    hm = plottetra(node, elem, *args, **kwargs)

    Plot 3D surface meshes.

    Parameters:
        node: (N, 3) or (N, 4) array of node coordinates (last column optional for color).
        elem: (M, 4) or (M, 5) array of tetrahedra (last column optional for tags).
        args, kwargs: Additional plotting options passed to plotsurf.

    Returns:
        hm: list of plot handles.
    """

    from mpl_toolkits.mplot3d.art3d import Poly3DCollection
    from matplotlib.colors import Normalize

    # Save current RNG state
    rngstate = np.random.get_state()

    # Set deterministic seed for consistent coloring
    randseed = int("623F9A9E", 16) + COLOR_OFFSET

    if "ISO2MESH_RANDSEED" in globals():
        randseed = globals()["ISO2MESH_RANDSEED"]

    np.random.seed(randseed)

    ax = plt.gca()

    h = {"fig": [], "ax": [], "obj": []}
    h["fig"].append(plt.gcf())
    h["ax"].append(ax)

    if not "color" in kwargs and not "cmap" in kwargs:
        kwargs["cmap"] = plt.get_cmap("jet")

    polydata = []
    colormap = []

    if isinstance(elem, list):
        elem = np.array(elem)

    if elem.shape[1] > 4:
        tag = elem[:, 4]  # 1-based -> column 5 in MATLAB
        types = np.unique(tag)
        for t in types:
            idx = np.where(tag == t)[0]
            face = volface(elem[idx, :4])[0]
            pdata, _ = plotasurf(node, face, *args, **kwargs)
            polydata.extend(pdata)
            colormap.extend(np.random.rand(1, 3).tolist() * len(pdata))
    else:
        face = volface(elem[:, :4])[0]
        polydata, colormap = plotasurf(node, face, *args, **kwargs)

    if "colormap" in locals() and len(colormap) > 0 and not "facecolors" in kwargs:
        kwargs["facecolors"] = colormap

    if "cmap" in kwargs and not "facecolors" in kwargs:
        node_values = node[:, 3] if node.shape[1] > 3 else node[:, 2]
        face_values = np.array([np.mean(node_values[f]) for f in elem[:, :4] - 1])
        norm = Normalize(vmin=face_values.min(), vmax=face_values.max())
        kwargs["facecolors"] = kwargs["cmap"](norm(face_values))

    if not "linewidth" in kwargs:
        kwargs["linewidth"] = 0.3

    patch = Poly3DCollection(polydata, edgecolors="k", **kwargs)
    ax.add_collection3d(patch)
    _autoscale_3d(ax, node)

    h["obj"].append(patch)

    # Restore RNG state
    np.random.set_state(rngstate)

    # Return handle if needed
    return h


# _________________________________________________________________________________________________________


def plotedges(node, edges, *args, **kwargs):
    """
    Plot a 3D polyline or closed loop (1D manifold).

    Parameters
    ----------
    node : ndarray (N, 3 or 4)
        Node coordinates. If a 4th column is present, it can represent color.
    edges : ndarray or list
        Either a 2-column edge list or a 1D list/array of node indices separated by NaN.
    *args : list
        Additional plotting options (passed to matplotlib).
    iso2mesh_randseed : int, optional
        Random seed for color generation (used for loops).

    Returns
    -------
    hh : list
        Handles to plotted elements.
    """
    hh = {"fig": [], "ax": [], "obj": []}
    edges = np.asarray(edges, order="F")  # Flatten in F order if needed

    if edges.size == 0:
        return hh

    rng_state = np.random.get_state()

    ax = plt.gca()

    hh["fig"].append(plt.gcf())
    hh["ax"].append(ax)

    if edges.ndim == 1 or edges.shape[1] == 1:
        # Loop: NaN-separated index list
        randseed = int("623F9A9E", 16) + COLOR_OFFSET
        if "iso2mesh_randseed" in kwargs:
            randseed = kwargs["iso2mesh_randseed"]
        np.random.seed(randseed)

        loops = edges.flatten(order="F")
        if not np.isnan(loops[-1]):
            loops = np.append(loops, np.nan)

        seg = np.where(np.isnan(loops))[0]
        seghead = 0

        for i in seg:
            segment = loops[seghead:i]
            segment = segment.astype(int) - 1  # 1-based to 0-based
            if segment.size > 1:
                (h,) = plt.plot(
                    node[segment, 0],
                    node[segment, 1],
                    node[segment, 2] if node.shape[1] >= 3 else None,
                    color=np.random.rand(
                        3,
                    ),
                    *args,
                    **kwargs,
                )
                hh["obj"].append(h)
            seghead = i + 1
    else:
        from mpl_toolkits.mplot3d.art3d import Line3DCollection

        # Edge list: each row connects 2 nodes
        edges = edges.astype(int) - 1  # 1-based to 0-based

        if node.shape[1] >= 3:
            segments = [[node[start], node[end]] for start, end in edges]
            h = Line3DCollection(segments, **kwargs)
            ax.add_collection3d(h)
            _autoscale_3d(ax, node)

        else:
            x = node[:, 0].flatten()
            y = node[:, 1].flatten()
            h = plt.plot(x[edges.T], y[edges.T], *args, **kwargs)

        hh["obj"].append(h)

    np.random.set_state(rng_state)
    return hh


# _________________________________________________________________________________________________________


def plotmesh(node, *args, **kwargs):
    """
    handles = plotmesh(node, face, elem, selector, ...)
    Plot surface and volumetric meshes in 3D.
    Converts 1-based MATLAB indices in `face` and `elem` to 0-based.
    Supports optional selector strings and stylistic options.
    """

    selector = None
    opt = []
    face = None
    elem = None
    node = np.array(node)

    # Parse inputs: detect selector strings, face/elem arrays, opts
    for i, a in enumerate(args):
        if isinstance(a, str):
            if any(c in a for c in "<>=&|") and any(c in a for c in "xyzXYZ"):
                selector = a
                opt = list(args[i + 1 :])
                break
            else:
                opt = list(args[i:])
                break
        else:
            if i == 0:
                if isinstance(a, list) or (
                    isinstance(a, np.ndarray) and a.ndim == 2 and a.shape[1] < 4
                ):
                    face = a
                elif isinstance(a, np.ndarray) and a.ndim == 2 and a.shape[1] in (4, 5):
                    uniq = np.unique(a[:, 3])
                    counts = np.bincount(a[:, 3].astype(int))
                    if len(uniq) == 1 or np.any(counts > 50):
                        face = a
                    else:
                        elem = a
                else:
                    elem = a
            elif i == 1:
                face = args[0]
                elem = a

    extraarg = {}
    if "hold" in kwargs:
        extraarg["hold"] = kwargs["hold"]

    ax = _createaxis(True, *args, **kwargs)

    handles = {"fig": [], "ax": [], "obj": []}
    handles["fig"].append(plt.gcf())
    handles["ax"].append(ax)

    for extraopt in ["hold", "parent", "subplot"]:
        if extraopt in kwargs:
            del kwargs[extraopt]

    # Plot points if no face/elem
    if face is None and elem is None:
        x, y, z = node[:, 0], node[:, 1], node[:, 2]
        idx = (
            np.where(eval(selector, {"x": x, "y": y, "z": z}))[0]
            if selector
            else slice(None)
        )
        if getattr(idx, "size", None) == 0:
            print("Warning: nothing to plot")
            return None
        (h,) = ax.plot(x[idx], y[idx], z[idx], *opt, **kwargs)
        handles["obj"].append(h)
        _autoscale_3d(ax, node)

    # Plot surface mesh
    if face is not None:
        if isinstance(face, list):
            handles = plotsurf(node, face, opt, *args, **kwargs)
        else:
            c0 = meshcentroid(node[:, :3], face[:, :3])
            x, y, z = c0[:, 0], c0[:, 1], c0[:, 2]
            idx = (
                np.where(eval(selector, {"x": x, "y": y, "z": z}))[0]
                if selector
                else slice(None)
            )
            if getattr(idx, "size", None) == 0:
                print("Warning: nothing to plot")
                return None
            handles = plotsurf(node, face[idx, :], opt, *args, **kwargs)

    # Plot tetrahedral mesh
    if elem is not None:
        c0 = meshcentroid(node[:, :3], elem[:, :4])
        x, y, z = c0[:, 0], c0[:, 1], c0[:, 2]
        idx = (
            np.where(eval(selector, {"x": x, "y": y, "z": z}))[0]
            if selector
            else slice(None)
        )
        if getattr(idx, "size", None) == 0:
            print("Warning: nothing to plot")
            return None
        if "alpha" not in kwargs and selector:
            faceselect, elemid = volface(elem[idx, :4])
            if elem.shape[1] > 4:
                faceselect = np.hstack(
                    (faceselect, elem[idx[elemid - 1], 4].reshape(-1, 1))
                )
            handles = plotsurf(node, faceselect, opt, *args, **kwargs)
        else:
            handles = plottetra(node, elem[idx, :], opt, *args, **kwargs)

    if not "hold" in extraarg or not extraarg["hold"] or extraarg["hold"] == "off":
        plt.draw()
        plt.show(block=False)

    return handles


def _autoscale_3d(ax, points):
    x, y, z = points[:, 0], points[:, 1], points[:, 2]
    boxas = np.array([x.max() - x.min(), y.max() - y.min(), z.max() - z.min()])
    if boxas[0] > 0:
        ax.set_xlim([x.min(), x.max()])
    if boxas[1] > 0:
        ax.set_ylim([y.min(), y.max()])
    if boxas[2] > 0:
        ax.set_zlim([z.min(), z.max()])
    if np.all(boxas > 0):
        ax.set_box_aspect(boxas)


def _createaxis(*args, **kwargs):
    subplotid = kwargs.get("subplot", 111)
    docreate = True
    fig = None

    if "parent" in kwargs:
        hh = kwargs["parent"]
        if isinstance(hh, dict):
            fig = hh["fig"][0]
            ax = hh["ax"][-1]
        elif isinstance(hh, list):
            ax = hh[-1]
        if "subplot" in kwargs and fig:
            ax = fig.add_subplot(subplotid, projection="3d")
    elif not docreate and len(plt.get_fignums()) > 0 and len(plt.gcf().axes) > 0:
        if not fig:
            fig = plt.gcf()
        ax = fig.axes[-1]
    else:
        if docreate and not fig:
            fig = plt.figure()
        ax = fig.add_subplot(subplotid, projection="3d")

    if ax.name != "3d":
        fig = plt.figure()  # Create a new figure
        ax = fig.add_subplot(subplotid, projection="3d")

    return ax


class plotvolume:
    """
    Interactive 3D volume slice viewer - Python port of MATLAB slice3i and mcxplotvol

    Display slices from a volume in 3-D with interactive controls.

    Vol is either a scalar or RGB volume, e.g. N x M x K or N x M x K x 3.
    This Python implementation uses matplotlib for 3D rendering with optimized
    single PatchCollection for better performance and no culling issues.

    Features:
    - Interactive slice dragging with accurate ray-casting
    - 3D rotation by clicking/holding the middle mouse button
    - Adjust colormap levels by dragging with the right mouse button
    - Support for 4-D data (up/down keys to change frames)
    - Subsampling for performance optimization
    - MATLAB-compatible interface

    Controls:
    - Left mouse: Click and drag to move slice positions
    - Middle mouse: Click and drag to rotate 3D view
    - Right mouse: Drag to adjust colormap levels
    - Sliders: Precise slice positioning
    - Up/Down keys: Navigate frames (4D volumes)

    Quality Settings:
    - subsample=1: Highest quality (slow)
    - subsample=2: High quality (recommended)
    - subsample=4: Fast rendering
    - subsample=8: Fastest rendering

    Example:

    import numpy as np
    vol = create_demo_volume()
    viewer = plotvolume(vol, subsample=2)
    viewer.setslice(0, 32)  # X-slice at position 32
    viewer.setslice(1, 32)  # Y-slice at position 32
    viewer.setslice(2, 14)  # Z-slice at position 14
    viewer.show()

    Original MATLAB implementation by:
    Author: Anders Brun, anders@cb.uu.se (2009)
    Modified by: Qianqian Fang, q.fang at neu.edu

    Python port features:
    - Ray-casting for accurate slice selection
    - Single PatchCollection for efficient rendering
    - Subsampling for performance optimization
    - Modern matplotlib integration
    """

    def __init__(
        self, vol, I2X=None, figsize=(12, 8), subsample=1, colormap="viridis", **kwargs
    ):
        """
        Initialize the interactive volume slicer

        Parameters:
        vol: 3D or 4D numpy array (N x M x K or N x M x K x 3/frames)
        I2X: 4x4 transformation matrix from volume coords to world coords
        figsize: figure size tuple
        subsample: subsampling factor for performance (1=highest quality, 4=faster)
        """
        self.vol = vol
        self.subsample = subsample  # Store subsampling factor

        # Disable I2X transformation - use identity matrix
        self.I2X = np.eye(4, dtype=float)

        # Handle different volume types
        self.current_vol = vol
        self.frame = 0
        self.params = kwargs

        if vol.ndim == 4 and vol.shape[3] > 3:
            # Multi-frame volume
            self.current_vol = vol[:, :, :, 0]
            self.is_multiframe = True
        else:
            self.is_multiframe = False

        # Initialize slice positions
        self.slice_positions = {
            0: self.current_vol.shape[0] // 2,  # i dimension (x-slice)
            1: self.current_vol.shape[1] // 2,  # j dimension (y-slice)
            2: self.current_vol.shape[2] // 2,  # k dimension (z-slice)
        }

        # GUI state for interaction
        self.gui_state = {
            "dragging": False,
            "rotating": False,
            "adjusting_colormap": False,
            "start_ray": None,
            "start_idx": None,
            "start_mouse_pos": None,
            "active_slicedim": None,
            "colormap_levels": 256,
            "rotation_enabled": False,
            "last_mouse_pos": None,
        }

        # Single patch collection for all slices
        self.patch_collection = None
        self.colormap_name = colormap

        self.setup_figure(figsize)
        self.setup_controls()
        self.update_patches()

    def setup_figure(self, figsize):
        """Setup the main figure and 3D axis"""
        # Close any existing figures to prevent multiple windows
        plt.close("all")

        self.fig = plt.figure(figsize=figsize)
        self.ax = self.fig.add_subplot(111, projection="3d")

        # Set initial view similar to MATLAB example
        self.ax.view_init(elev=45, azim=30)
        self.ax.set_xlabel("X")
        self.ax.set_ylabel("Y")
        self.ax.set_zlabel("Z")

        # Disable default mouse rotation
        self.ax.mouse_init(rotate_btn=None, zoom_btn=None)

        # Set equal aspect ratio
        self.set_equal_aspect()

        # Connect mouse and keyboard events
        self.fig.canvas.mpl_connect("button_press_event", self.on_mouse_press)
        self.fig.canvas.mpl_connect("button_release_event", self.on_mouse_release)
        self.fig.canvas.mpl_connect("motion_notify_event", self.on_mouse_move)
        self.fig.canvas.mpl_connect("key_press_event", self.on_key_press)

    def setup_controls(self):
        """Setup control widgets with proper spacing to avoid overlap"""
        # Create even more space for controls
        from matplotlib.widgets import Slider

        plt.subplots_adjust(bottom=0.3)

        # X-slice slider - leftmost position, smaller width
        ax_x = plt.axes([0.05, 0.18, 0.20, 0.03])
        self.slider_x = Slider(
            ax_x,
            "X-slice",
            0,
            self.current_vol.shape[0] - 1,
            valinit=self.slice_positions[0],
            valfmt="%d",
        )
        self.slider_x.on_changed(self.update_x_slice)

        # Y-slice slider - center position with gap
        ax_y = plt.axes([0.35, 0.18, 0.20, 0.03])
        self.slider_y = Slider(
            ax_y,
            "Y-slice",
            0,
            self.current_vol.shape[1] - 1,
            valinit=self.slice_positions[1],
            valfmt="%d",
        )
        self.slider_y.on_changed(self.update_y_slice)

        # Z-slice slider - rightmost position with gap
        ax_z = plt.axes([0.66, 0.18, 0.20, 0.03])
        self.slider_z = Slider(
            ax_z,
            "Z-slice",
            0,
            self.current_vol.shape[2] - 1,
            valinit=self.slice_positions[2],
            valfmt="%d",
        )
        self.slider_z.on_changed(self.update_z_slice)

        # Colormap levels slider - full width, positioned lower with more gap
        ax_cmap = plt.axes([0.05, 0.10, 0.64, 0.03])
        self.colormap_slider = Slider(
            ax_cmap, "Color Levels", 8, 256, valinit=256, valfmt="%d"
        )
        self.colormap_slider.on_changed(self.update_colormap_levels)

        # Frame slider for 4D volumes - positioned at bottom with more gap
        if self.is_multiframe:
            ax_frame = plt.axes([0.05, 0.02, 0.64, 0.03])
            self.frame_slider = Slider(
                ax_frame, "Frame", 0, self.vol.shape[3] - 1, valinit=0, valfmt="%d"
            )
            self.frame_slider.on_changed(self.update_frame)

    def set_equal_aspect(self):
        """Set 1:1:1 aspect ratio with correct axis bounds (no transformation)"""
        # Get volume shape
        vol_shape = np.array(self.current_vol.shape)

        # Use volume indices directly (no transformation)
        padding = 2  # Small padding in voxel units

        x_min, x_max = -padding, vol_shape[0] + padding  # 0 to 64
        y_min, y_max = -padding, vol_shape[1] + padding  # 0 to 64
        z_min, z_max = -padding, vol_shape[2] + padding  # 0 to 27

        # Set the axis limits
        self.ax.set_xlim(x_min, x_max)
        self.ax.set_ylim(y_min, y_max)
        self.ax.set_zlim(z_min, z_max)

        # Set 1:1:1 aspect ratio - equal scaling for all axes
        self.ax.set_box_aspect([1, 1, 1])

        # Force matplotlib to respect the aspect ratio
        try:
            self.ax.set_aspect("equal")
        except:
            pass  # Some matplotlib versions don't support this for 3D

    def create_slice_polygons(self, slicedim, sliceidx):
        """
        Create polygon vertices for a slice with optional subsampling
        Returns list of 4-vertex polygons (quads) and corresponding colors
        """
        vol = self.current_vol
        sliceidx = int(np.clip(sliceidx, 0, vol.shape[slicedim] - 1))
        s = self.subsample  # Subsampling factor

        if slicedim == 2:  # k dimension (z-slice)
            # Create grid of quads
            slice_data = vol[:, :, sliceidx]
            ni, nj = slice_data.shape

            polygons = []
            colors = []

            for i in range(0, ni - s, s):
                for j in range(0, nj - s, s):
                    # Four corners of the quad in volume coordinates
                    corners_vol = np.array(
                        [
                            [i, j, sliceidx],
                            [i + s, j, sliceidx],
                            [i + s, j + s, sliceidx],
                            [i, j + s, sliceidx],
                        ]
                    )

                    polygons.append(corners_vol)

                    # Use average color of the subsampled region
                    i_end = min(i + s, ni)
                    j_end = min(j + s, nj)
                    region = slice_data[i:i_end, j:j_end]
                    quad_color = np.mean(region)
                    colors.append(quad_color)

        elif slicedim == 1:  # j dimension (y-slice)
            slice_data = vol[:, sliceidx, :]
            ni, nk = slice_data.shape

            polygons = []
            colors = []

            for i in range(0, ni - s, s):
                for k in range(0, nk - s, s):
                    corners_vol = np.array(
                        [
                            [i, sliceidx, k],
                            [i + s, sliceidx, k],
                            [i + s, sliceidx, k + s],
                            [i, sliceidx, k + s],
                        ]
                    )

                    polygons.append(corners_vol)

                    i_end = min(i + s, ni)
                    k_end = min(k + s, nk)
                    region = slice_data[i:i_end, k:k_end]
                    quad_color = np.mean(region)
                    colors.append(quad_color)

        elif slicedim == 0:  # i dimension (x-slice)
            slice_data = vol[sliceidx, :, :]
            nj, nk = slice_data.shape

            polygons = []
            colors = []

            for j in range(0, nj - s, s):
                for k in range(0, nk - s, s):
                    corners_vol = np.array(
                        [
                            [sliceidx, j, k],
                            [sliceidx, j + s, k],
                            [sliceidx, j + s, k + s],
                            [sliceidx, j, k + s],
                        ]
                    )

                    polygons.append(corners_vol)

                    j_end = min(j + s, nj)
                    k_end = min(k + s, nk)
                    region = slice_data[j:j_end, k:k_end]
                    quad_color = np.mean(region)
                    colors.append(quad_color)

        return polygons, colors

    def update_patches(self):
        """Update the single patch collection with all slice polygons"""
        # Remove existing patch collection
        from mpl_toolkits.mplot3d.art3d import Poly3DCollection
        from matplotlib.colors import Normalize
        import matplotlib.cm as cm

        if self.patch_collection is not None:
            self.patch_collection.remove()

        # Collect all polygons and colors from all active slices
        all_polygons = []
        all_colors = []

        # Add polygons from each slice dimension
        for slicedim in [0, 1, 2]:
            polygons, colors = self.create_slice_polygons(
                slicedim, self.slice_positions[slicedim]
            )
            all_polygons.extend(polygons)
            all_colors.extend(colors)

        if all_polygons:
            # Normalize colors
            all_colors = np.array(all_colors)
            if all_colors.max() > 0:
                norm = Normalize(vmin=all_colors.min(), vmax=all_colors.max())
                normalized_colors = norm(all_colors)
            else:
                normalized_colors = all_colors

            if "alpha" not in self.params:
                self.params["alpha"] = 1.0
            if "edgecolors" not in self.params:
                self.params["edgecolors"] = "none"
            if "linewidths" not in self.params:
                self.params["linewidths"] = 0.0
            if "shade" not in self.params:
                self.params["shade"] = False
            if "antialiased" not in self.params:
                self.params["antialiased"] = False
            if "rasterized" not in self.params:
                self.params["rasterized"] = True

            # Create single Poly3DCollection
            self.patch_collection = Poly3DCollection(
                all_polygons,
                facecolors=cm.get_cmap(self.colormap_name)(normalized_colors),
                **self.params,
            )

            self.ax.add_collection3d(self.patch_collection)

        self.fig.canvas.draw_idle()

    def get_mouse_ray(self, event):
        """Get the 3D ray (origin and direction) from a mouse click on 3D axes."""
        if event.xdata is None or event.ydata is None:
            return None, None

        # Step 1: Get renderer and projection matrix
        ax = self.ax
        fig = self.fig
        proj = ax.get_proj()

        # Step 2: Get axes bbox instead of figure bbox
        ax_bbox = ax.get_window_extent()

        # Convert event coordinates to axes-relative coordinates
        x_screen = event.x - ax_bbox.x0
        y_screen = event.y - ax_bbox.y0

        # Invert Y coordinate relative to axes
        y_screen = ax_bbox.height - y_screen

        def inv_proj(x, y, z):
            """Invert projection from screen coords to data coords"""
            # Normalize screen coordinates to axes bbox, then to [-1,1]
            w, h = ax_bbox.width, ax_bbox.height
            x_ndc = 2 * x / w - 1
            y_ndc = 2 * y / h - 1
            z_ndc = 2 * z - 1
            ndc = np.array([x_ndc, y_ndc, z_ndc, 1.0])
            try:
                inv = np.linalg.inv(proj)
            except np.linalg.LinAlgError:
                return None
            world = inv @ ndc
            if abs(world[3]) < 1e-10:
                return None
            world /= world[3]
            return world[:3]

        # Step 3: Compute world coords at near and far planes
        p_near = inv_proj(x_screen, y_screen, 0.0)
        p_far = inv_proj(x_screen, y_screen, 1.0)

        if p_near is None or p_far is None:
            return None, None

        origin = p_near
        direction = p_far - p_near
        direction_length = np.linalg.norm(direction)

        if direction_length < 1e-10:
            return None, None

        direction /= direction_length

        return origin, direction

    def get_slice_rectangle_bounds(self, slicedim, sliceidx):
        """
        Get the rectangle boundary coordinates
        Returns 4 corner points of the rectangular slice in volume coordinates
        """
        vol_shape = self.current_vol.shape

        if slicedim == 0:  # X-slice (cutting perpendicular to X-axis, showing YZ plane)
            # Fixed X coordinate, varying Y and Z
            corners = np.array(
                [
                    [sliceidx, 0, 0],  # (X, Y_min, Z_min)
                    [sliceidx, vol_shape[1] - 1, 0],  # (X, Y_max, Z_min)
                    [sliceidx, vol_shape[1] - 1, vol_shape[2] - 1],  # (X, Y_max, Z_max)
                    [sliceidx, 0, vol_shape[2] - 1],  # (X, Y_min, Z_max)
                ]
            )

        elif (
            slicedim == 1
        ):  # Y-slice (cutting perpendicular to Y-axis, showing XZ plane)
            # Fixed Y coordinate, varying X and Z
            corners = np.array(
                [
                    [0, sliceidx, 0],  # (X_min, Y, Z_min)
                    [vol_shape[0] - 1, sliceidx, 0],  # (X_max, Y, Z_min)
                    [vol_shape[0] - 1, sliceidx, vol_shape[2] - 1],  # (X_max, Y, Z_max)
                    [0, sliceidx, vol_shape[2] - 1],  # (X_min, Y, Z_max)
                ]
            )

        else:  # slicedim == 2, Z-slice (cutting perpendicular to Z-axis, showing XY plane)
            # Fixed Z coordinate, varying X and Y
            corners = np.array(
                [
                    [0, 0, sliceidx],  # (X_min, Y_min, Z)
                    [vol_shape[0] - 1, 0, sliceidx],  # (X_max, Y_min, Z)
                    [vol_shape[0] - 1, vol_shape[1] - 1, sliceidx],  # (X_max, Y_max, Z)
                    [0, vol_shape[1] - 1, sliceidx],  # (X_min, Y_max, Z)
                ]
            )

        return corners

    def point_in_rectangle(self, point, rectangle_corners):
        """
        Test if a 3D point lies within a rectangular slice boundary
        Simplified and more robust version using 2D projection
        """
        # Get rectangle vectors
        p0, p1, p2, p3 = rectangle_corners

        # Create two edge vectors from p0
        edge1 = p1 - p0  # First edge
        edge2 = p3 - p0  # Adjacent edge

        # Vector from p0 to test point
        point_vec = point - p0

        # Project the point onto the rectangle's local coordinate system
        # Use dot products to get coordinates in the rectangle's basis
        edge1_len_sq = np.dot(edge1, edge1)
        edge2_len_sq = np.dot(edge2, edge2)

        if edge1_len_sq < 1e-10 or edge2_len_sq < 1e-10:
            return False  # Degenerate rectangle

        # Get local coordinates (u, v) in rectangle space
        u = np.dot(point_vec, edge1) / edge1_len_sq
        v = np.dot(point_vec, edge2) / edge2_len_sq

        # Point is inside rectangle if 0 <= u <= 1 and 0 <= v <= 1
        return 0 <= u <= 1 and 0 <= v <= 1

    def ray_rectangle_intersection(self, ray_origin, ray_direction, rectangle_corners):
        """
        Calculate intersection of ray with rectangular slice boundary
        Improved version with better debugging
        """
        # Get rectangle corners
        p0, p1, p2, p3 = rectangle_corners

        # Calculate plane normal using cross product of two edges
        edge1 = p1 - p0
        edge2 = p3 - p0
        plane_normal = np.cross(edge1, edge2)

        # Check for degenerate rectangle
        normal_length = np.linalg.norm(plane_normal)
        if normal_length < 1e-10:
            return None, None

        plane_normal = plane_normal / normal_length

        # Ray-plane intersection using parametric form
        # Ray: P = ray_origin + t * ray_direction
        # Plane: (P - p0) · normal = 0
        denominator = np.dot(ray_direction, plane_normal)

        if abs(denominator) < 1e-10:  # Ray parallel to plane
            return None, None

        # Calculate intersection parameter t
        t = np.dot(p0 - ray_origin, plane_normal) / denominator

        if t < 0:  # Intersection behind camera
            return None, None

        # Calculate intersection point
        intersection = ray_origin + t * ray_direction

        # Check if intersection is within rectangle bounds
        if self.point_in_rectangle(intersection, rectangle_corners):
            distance = np.linalg.norm(intersection - ray_origin)
            return intersection, distance
        else:
            return None, None

    def detect_clicked_slice_raycast(self, event):
        """
        Detect which slice was clicked using proper ray-rectangle intersection
        Only considers actual rectangular slice boundaries
        """
        if not event.inaxes == self.ax:
            return 2

        # Get ray from camera through mouse
        ray_origin, ray_direction = self.get_mouse_ray(event)
        if ray_origin is None:
            return self.detect_slice_from_view_angle()

        # Test intersection with each slice rectangle
        valid_intersections = []

        for slicedim in [0, 1, 2]:
            sliceidx = self.slice_positions[slicedim]

            # Get the actual rectangular bounds of this slice
            rectangle_corners = self.get_slice_rectangle_bounds(slicedim, sliceidx)

            # Calculate ray-rectangle intersection
            intersection, distance = self.ray_rectangle_intersection(
                ray_origin, ray_direction, rectangle_corners
            )

            if intersection is not None and distance is not None:
                valid_intersections.append((slicedim, distance, intersection))

        # Return the closest valid intersection
        if valid_intersections:
            # Sort by distance (closest first)
            valid_intersections.sort(key=lambda x: x[1])
            closest_slice = valid_intersections[0][0]
            return closest_slice

        # Fallback if no intersections found
        return self.detect_slice_from_view_angle()

    def detect_slice_from_view_angle(self):
        """
        Fallback method: detect slice based on current view angles
        More reliable when 3D coordinates are not available
        """
        try:
            elev = self.ax.elev
            azim = self.ax.azim

            # Normalize angles
            azim = azim % 360
            elev = max(-90, min(90, elev))

            # Simple heuristic based on view angles
            # This determines which slice is most "facing" the camera

            if abs(elev) > 60:  # Looking from top or bottom
                return 2  # Z-slice
            elif 30 <= azim <= 150 or 210 <= azim <= 330:  # Looking from sides
                return 0  # X-slice
            else:  # Looking from front or back
                return 1  # Y-slice

        except (AttributeError, ValueError):
            # Ultimate fallback
            return 2  # Default to Z-slice

    def update_x_slice(self, val):
        """Update X-slice from slider"""
        self.slice_positions[0] = int(val)
        self.update_patches()

    def update_y_slice(self, val):
        """Update Y-slice from slider"""
        self.slice_positions[1] = int(val)
        self.update_patches()

    def update_z_slice(self, val):
        """Update Z-slice from slider"""
        self.slice_positions[2] = int(val)
        self.update_patches()

    def update_colormap_levels(self, val):
        """Update colormap levels"""
        levels = int(val)
        self.gui_state["colormap_levels"] = levels
        # For simplicity, we'll just update the patches
        self.update_patches()

    def update_frame(self, val):
        """Update frame for 4D volumes"""
        if self.is_multiframe:
            self.frame = int(val)
            self.current_vol = self.vol[:, :, :, self.frame]
            self.update_patches()

    def on_mouse_press(self, event):
        """Handle mouse press events"""
        if event.inaxes != self.ax:
            return

        if event.button == 2:  # Middle mouse - toggle rotation mode
            self.gui_state["rotation_enabled"] = True
            self.gui_state["rotating"] = True
            self.gui_state["last_mouse_pos"] = (
                event.x,
                event.y,
            )  # Use pixel coordinates
            # Change cursor to indicate rotation mode
            try:
                self.fig.canvas.set_cursor(2)  # Hand cursor
            except:
                pass  # Ignore cursor errors
            return

        elif event.button == 3:  # Right mouse - colormap adjustment
            if not self.gui_state[
                "rotating"
            ]:  # Don't allow colormap adjustment during rotation
                self.gui_state["adjusting_colormap"] = True
                self.gui_state["start_mouse_pos"] = (event.x, event.y)
            return

        elif event.button == 1:  # Left mouse - slice dragging
            if not self.gui_state[
                "rotating"
            ]:  # Don't allow slice dragging during rotation
                self.gui_state["dragging"] = True
                self.gui_state["start_mouse_pos"] = (event.x, event.y)

                # Use ray-casting to detect which slice to drag
                detected_slice = self.detect_clicked_slice_raycast(event)
                self.gui_state["active_slicedim"] = detected_slice
                self.gui_state["start_idx"] = self.slice_positions[detected_slice]

                # Change cursor to indicate dragging mode
                try:
                    self.fig.canvas.set_cursor(4)  # Move cursor
                except:
                    pass  # Ignore cursor errors

    def on_mouse_release(self, event):
        """Handle mouse release events"""
        if event.button == 2:  # Middle mouse release - disable rotation
            self.gui_state["rotating"] = False
            self.gui_state["rotation_enabled"] = False
            self.gui_state["last_mouse_pos"] = None
            try:
                self.fig.canvas.set_cursor(1)  # Default cursor
            except:
                pass

        elif event.button == 3:  # Right mouse release
            self.gui_state["adjusting_colormap"] = False
            self.gui_state["start_mouse_pos"] = None
            try:
                self.fig.canvas.set_cursor(1)  # Default cursor
            except:
                pass

        elif event.button == 1:  # Left mouse release
            self.gui_state["dragging"] = False
            self.gui_state["start_mouse_pos"] = None
            self.gui_state["start_idx"] = None
            self.gui_state["active_slicedim"] = None
            try:
                self.fig.canvas.set_cursor(1)  # Default cursor
            except:
                pass

    def on_mouse_move(self, event):
        """Handle mouse move events"""
        if not event.inaxes == self.ax:
            return

        if self.gui_state["rotating"] and self.gui_state["last_mouse_pos"]:
            # Manual 3D rotation (middle mouse drag)
            if event.x is not None and event.y is not None:
                last_x, last_y = self.gui_state["last_mouse_pos"]
                dx = event.x - last_x
                dy = event.y - last_y

                # Get current view angles
                elev = self.ax.elev
                azim = self.ax.azim

                # Update angles based on mouse movement (pixel coordinates)
                # Scale factors for sensitivity
                azim_sensitivity = 0.5
                elev_sensitivity = 0.5

                new_azim = azim + dx * azim_sensitivity
                new_elev = elev - dy * elev_sensitivity  # Invert Y for natural rotation

                # Clamp elevation to reasonable bounds
                new_elev = max(-90, min(90, new_elev))

                # Apply new view
                self.ax.view_init(elev=new_elev, azim=new_azim)

                # Update axis limits to keep slices properly visible after rotation
                self.set_equal_aspect()

                self.gui_state["last_mouse_pos"] = (event.x, event.y)
                self.fig.canvas.draw_idle()

        elif self.gui_state["adjusting_colormap"] and self.gui_state["start_mouse_pos"]:
            # Colormap adjustment (right mouse drag) - only if not rotating
            start_x, start_y = self.gui_state["start_mouse_pos"]
            delta_y = event.y - start_y

            # Scale the delta to colormap levels
            fig_height = self.fig.get_size_inches()[1] * self.fig.dpi
            normalized_delta = delta_y / fig_height

            new_levels = int(256 + normalized_delta * 256 * 4)
            new_levels = max(8, min(256, new_levels))

            self.colormap_slider.set_val(new_levels)

        elif self.gui_state["dragging"] and self.gui_state["start_mouse_pos"]:
            # Slice dragging (left mouse drag) - only if not rotating
            start_x, start_y = self.gui_state["start_mouse_pos"]
            delta_x = event.x - start_x
            delta_y = event.y - start_y

            # Convert mouse movement to slice index change
            # Note: Mouse Y coordinates are inverted (down = positive Y)
            # For intuitive control, we want:
            # - Mouse down = slice index decreases (for Z-slice: move down in volume)
            # - Mouse up = slice index increases (for Z-slice: move up in volume)
            # - Mouse right = slice index increases
            # - Mouse left = slice index decreases

            slicedim = self.gui_state.get("active_slicedim", 2)

            if slicedim == 2:  # Z-slice - use Y movement (inverted)
                movement = -delta_y  # Invert Y movement for natural control
            elif slicedim == 1:  # Y-slice - use Y movement (inverted)
                movement = -delta_y  # Invert Y movement
            else:  # X-slice - use X movement
                movement = delta_x

            # Scale movement to slice indices
            max_slices = self.current_vol.shape[slicedim]
            slice_delta = movement / 20.0  # Adjust sensitivity
            new_idx = self.gui_state["start_idx"] + slice_delta
            new_idx = max(0, min(max_slices - 1, int(new_idx)))

            if new_idx != self.slice_positions[slicedim]:
                self.slice_positions[slicedim] = new_idx

                # Update corresponding slider
                if slicedim == 0:
                    self.slider_x.set_val(new_idx)
                elif slicedim == 1:
                    self.slider_y.set_val(new_idx)
                elif slicedim == 2:
                    self.slider_z.set_val(new_idx)

    def on_key_press(self, event):
        """Handle key press events"""
        if self.is_multiframe:
            if event.key == "up":
                new_frame = min(self.frame + 1, self.vol.shape[3] - 1)
                self.frame_slider.set_val(new_frame)
            elif event.key == "down":
                new_frame = max(self.frame - 1, 0)
                self.frame_slider.set_val(new_frame)

    def set_quality(self, subsample):
        """
        Set rendering quality by changing subsample factor

        Parameters:
        subsample: 1=highest quality (slow), 2=high quality, 4=fast, 8=fastest
        """
        self.subsample = max(1, int(subsample))
        self.update_patches()

    def setslice(self, slicedim, sliceidx):
        """
        Set a specific slice position (MATLAB-compatible interface)
        slicedim: 0, 1, or 2 for i, j, k dimensions
        sliceidx: slice index
        """
        self.slice_positions[slicedim] = int(sliceidx)
        self.update_patches()

        # Update corresponding slider
        if slicedim == 0:
            self.slider_x.set_val(sliceidx)
        elif slicedim == 1:
            self.slider_y.set_val(sliceidx)
        elif slicedim == 2:
            self.slider_z.set_val(sliceidx)

    def show(self):
        """Display the interactive viewer"""
        # Don't call plt.show() here since the figure is already created and managed
        # Just make sure the figure is drawn
        self.fig.canvas.draw()
        plt.show()



"""@package docstring
Iso2Mesh for Python - Mesh data queries and manipulations

Copyright (c) 2024-2025 Qianqian Fang <q.fang at neu.edu>
"""
__all__ = [
    "volgrow",
    "volshrink",
    "volopen",
    "volclose",
    "fillholes3d",
    "thickenbinvol",
    "thinbinvol",
]

##====================================================================================
## dependent libraries
##====================================================================================

import numpy as np
from scipy import ndimage
from typing import Optional


##====================================================================================
## implementations
##====================================================================================


def validatemask(mask, ndim=3):
    # Create default mask if not provided or empty
    if mask is None or mask.size == 0:
        if ndim == 3:
            # Create 3D cross-shaped mask for 3D volumes
            mask = ndimage.generate_binary_structure(3, 1)
        else:
            # Create 2D cross-shaped mask for 2D images
            mask = ndimage.generate_binary_structure(2, 1)

    # Rotate mask by 180 degrees (equivalent to rot90(mask, 2) in MATLAB)
    if mask.ndim == 3:
        # For 3D arrays, rotate around all axes
        mask = np.rot90(mask, 2, axes=(0, 1))
        mask = np.rot90(mask, 2, axes=(0, 2))
    else:
        # For 2D arrays, simple 180 degree rotation
        mask = np.rot90(mask, 2)

    return mask


def volgrow(
    vol: np.ndarray, layer: int = 1, mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Thickening a binary image or volume by a given pixel width
    This is similar to bwmorph(vol,'thicken',3) except
    this does it in both 2d and 3d

    Author: Qianqian Fang, <q.fang at neu.edu>
    Python version adapted from original MATLAB code

    Parameters:
    -----------
    vol : ndarray
        A volumetric binary image
    layer : int, optional
        Number of iterations for the thickening (default: 1)
    mask : ndarray, optional
        A 2D or 3D neighborhood mask (default: None, will create appropriate mask)

    Returns:
    --------
    newvol : ndarray
        The volume image after the thickening

    Notes:
    ------
    This function is part of iso2mesh toolbox (http://iso2mesh.sf.net)
    """

    mask = validatemask(mask, vol.ndim)

    # Convert vol to appropriate type for processing
    newvol = vol.astype(np.float32)

    # Perform iterative dilation using scipy's binary_dilation
    # which is more appropriate for binary morphological operations
    mask_bool = mask > 0

    # Use scipy's binary_dilation for proper binary morphological operation
    newvol = ndimage.binary_dilation(newvol > 0, structure=mask_bool, iterations=layer)

    # Convert back to double precision (equivalent to MATLAB's double())
    newvol = newvol.astype(np.float64)

    return newvol


def volshrink(
    vol: np.ndarray, layer: int = 1, mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Alternative implementation using scipy's binary_erosion for proper morphological thinning

    This version uses scipy's binary_erosion which is more mathematically appropriate
    for binary morphological thinning operations.

    Parameters:
    -----------
    vol : ndarray
        A volumetric binary image
    layer : int, optional
        Number of iterations for the thinning (default: 1)
    mask : ndarray, optional
        A 2D or 3D neighborhood mask (default: None, will create appropriate mask)

    Returns:
    --------
    newvol : ndarray
        The volume image after the thinning operations
    """

    mask = validatemask(mask, vol.ndim)

    # Convert input to binary
    newvol = vol != 0

    # Perform iterative binary erosion (morphological thinning)
    newvol = ndimage.binary_erosion(
        newvol, structure=mask, iterations=layer, border_value=1
    )

    # Convert back to double precision
    newvol = newvol.astype(np.float64)

    return newvol


def volclose(
    vol: np.ndarray, layer: int = 1, mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Alternative implementation using scipy's binary_closing for proper morphological closing

    This version uses scipy's optimized binary_closing operation which is more
    mathematically appropriate and efficient for morphological closing.

    Parameters:
    -----------
    vol : ndarray
        A volumetric binary image
    layer : int, optional
        Number of iterations for the closing (default: 1)
    mask : ndarray, optional
        A 2D or 3D neighborhood mask (default: None, will create appropriate mask)

    Returns:
    --------
    newvol : ndarray
        The volume image after closing
    """

    # Validate input
    if vol is None:
        raise ValueError("must provide a volume")

    mask = validatemask(mask, vol.ndim)

    # Convert input to binary
    newvol = vol != 0

    # Perform iterative binary closing
    newvol = ndimage.binary_closing(newvol, structure=mask, iterations=layer)

    # Convert back to double precision
    newvol = newvol.astype(np.float64)

    return newvol


def volopen(
    vol: np.ndarray, layer: int = 1, mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Alternative implementation using scipy's binary_closing for proper morphological closing

    This version uses scipy's optimized binary_closing operation which is more
    mathematically appropriate and efficient for morphological closing.

    Parameters:
    -----------
    vol : ndarray
        A volumetric binary image
    layer : int, optional
        Number of iterations for the closing (default: 1)
    mask : ndarray, optional
        A 2D or 3D neighborhood mask (default: None, will create appropriate mask)

    Returns:
    --------
    newvol : ndarray
        The volume image after closing
    """

    # Validate input
    if vol is None:
        raise ValueError("must provide a volume")

    mask = validatemask(mask, vol.ndim)

    # Convert input to binary
    newvol = vol != 0

    # Perform iterative binary closing
    newvol = ndimage.binary_opening(newvol, structure=mask, iterations=layer)

    # Convert back to double precision
    newvol = newvol.astype(np.float64)

    return newvol


def fillholes3d(
    img: np.ndarray, maxgap=None, mask: Optional[np.ndarray] = None
) -> np.ndarray:
    """
    Close a 3D image with the specified gap size and then fill the holes

    Author: Qianqian Fang, <q.fang at neu.edu>
    Python version adapted from original MATLAB code

    Parameters:
    -----------
    img : ndarray
        A 2D or 3D binary image
    maxgap : int, ndarray, list, tuple, or None
        If is a scalar, specify maximum gap size for image closing
        If a pair of coordinates, specify the seed position for floodfill
        If None, no initial closing operation is performed
    mask : ndarray, optional
        Neighborhood structure element for floodfilling (default: None)

    Returns:
    --------
    resimg : ndarray
        The image free of holes

    Notes:
    ------
    This function is part of iso2mesh toolbox (http://iso2mesh.sf.net)

    The function works in two phases:
    1. If maxgap is a scalar > 0, apply morphological closing to bridge small gaps
    2. Fill holes using either scipy's binary_fill_holes or custom flood-fill algorithm

    When maxgap is coordinates, it specifies seed points for flood-filling specific regions.
    """

    # Convert to binary and fill holes
    binary_img = img > 0
    if maxgap:
        binary_img = volclose(binary_img, maxgap, mask)

    filled = ndimage.binary_fill_holes(binary_img)

    # Convert back to float64
    resimg = filled.astype(np.float64)

    return resimg


##====================================================================================
## aliases
##====================================================================================

thickenbinvol = volgrow
thinbinvol = volshrink



"""pyiso2mesh - Iso2Mesh for Python
"""

from .geometry import (
    meshgrid5,
    meshgrid6,
    latticegrid,
    meshabox,
    meshacylinder,
    meshcylinders,
    meshcylinders,
    meshanellip,
    meshunitsphere,
    meshasphere,
    extrudecurve,
    extrudesurf,
)

from .plot import (
    plotsurf,
    plotasurf,
    plotmesh,
    plottetra,
    plotedges,
    plotvolume,
)

from .trait import (
    finddisconnsurf,
    surfedge,
    volface,
    extractloops,
    meshconn,
    nodevolume,
    elemvolume,
    neighborelem,
    layersurf,
    faceneighbors,
    edgeneighbors,
    maxsurf,
    flatsegment,
    surfplane,
    surfinterior,
    surfpart,
    surfseeds,
    meshquality,
    meshedge,
    meshface,
    surfacenorm,
    nodesurfnorm,
    uniqedges,
    uniqfaces,
    innersurf,
    advancefront,
    meshreorient,
    mesheuler,
    raytrace,
    elemfacecenter,
    barydualmesh,
    meshcentroid,
    highordertet,
    ismember_rows,
    ray2surf,
)

from .utils import (
    getexeext,
    fallbackexeext,
    mwpath,
    mcpath,
    deletemeshfile,
    rotatevec3d,
    rotmat2vec,
    varargin2struct,
    jsonopt,
    nargout,
)

from .io import (
    saveinr,
    saveoff,
    saveasc,
    savestl,
    savebinstl,
    readmedit,
    readtetgen,
    savesurfpoly,
    readoff,
)

from .core import (
    v2m,
    v2s,
    s2m,
    s2v,
    vol2mesh,
    vol2surf,
    surf2mesh,
    surf2volz,
    surf2vol,
    binsurface,
    cgalv2m,
    cgals2m,
    getintersecttri,
    vol2restrictedtri,
    fillsurf,
    surfvolume,
    insurface,
    remeshsurf,
    meshrefine,
    outersurf,
)

from .modify import (
    sms,
    smoothsurf,
    qmeshcut,
    meshcheckrepair,
    removedupelem,
    removedupnodes,
    removeisolatednode,
    removeisolatedsurf,
    surfaceclean,
    removeedgefaces,
    getintersecttri,
    delendelem,
    surfreorient,
    sortmesh,
    cart2sph,
    sortrows,
    mergemesh,
    mergesurf,
    surfboolean,
    meshresample,
    domeshsimplify,
)

from .line import (
    linextriangle,
    getplanefrom3pt,
    polylinelen,
    closestnode,
    polylineinterp,
    polylinesimplify,
)

from .raster import (
    m2v,
    mesh2vol,
    mesh2mask,
    barycentricgrid,
)

from .register import (
    affinemap,
    meshinterp,
    meshremap,
    proj2mesh,
    dist2surf,
    regpt2surf,
)

from .brain2mesh import (
    brain2mesh,
    brain1020,
    intriangulation,
    label2tpm,
    tpm2label,
)

from .volume import (
    volgrow,
    volshrink,
    volopen,
    volclose,
    fillholes3d,
    thickenbinvol,
    thinbinvol,
)

from jdata import *

__version__ = "0.4.2"
__all__ = [
    "advancefront",
    "barycentricgrid",
    "barydualmesh",
    "binsurface",
    "cart2sph",
    "cgals2m",
    "cgalv2m",
    "delendelem",
    "deletemeshfile",
    "domeshsimplify",
    "edgeneighbors",
    "elemfacecenter",
    "elemvolume",
    "extractloops",
    "extrudecurve",
    "extrudesurf",
    "faceneighbors",
    "fallbackexeext",
    "fillsurf",
    "finddisconnsurf",
    "flatsegment",
    "getexeext",
    "getintersecttri",
    "getintersecttri",
    "highordertet",
    "innersurf",
    "insurface",
    "jsonopt",
    "latticegrid",
    "layersurf",
    "m2v",
    "maxsurf",
    "mcpath",
    "mergemesh",
    "mergesurf",
    "mesh2mask",
    "mesh2vol",
    "meshabox",
    "meshacylinder",
    "meshcylinders",
    "meshanellip",
    "meshasphere",
    "meshcentroid",
    "meshcheckrepair",
    "meshconn",
    "meshcylinders",
    "meshedge",
    "mesheuler",
    "meshface",
    "meshgrid5",
    "meshgrid6",
    "meshquality",
    "meshrefine",
    "meshreorient",
    "meshresample",
    "meshunitsphere",
    "mwpath",
    "neighborelem",
    "nodesurfnorm",
    "nodevolume",
    "outersurf",
    "plotasurf",
    "plotedges",
    "plotmesh",
    "plotsurf",
    "plottetra",
    "plotvolume",
    "qmeshcut",
    "raytrace",
    "readmedit",
    "readoff",
    "readtetgen",
    "remeshsurf",
    "removedupelem",
    "removedupnodes",
    "removeedgefaces",
    "removeisolatednode",
    "removeisolatedsurf",
    "rotatevec3d",
    "rotmat2vec",
    "s2m",
    "s2v",
    "saveasc",
    "savebinstl",
    "saveinr",
    "saveoff",
    "savestl",
    "savesurfpoly",
    "smoothsurf",
    "sms",
    "sortmesh",
    "sortrows",
    "surf2mesh",
    "surf2vol",
    "surf2volz",
    "surfaceclean",
    "surfacenorm",
    "surfboolean",
    "surfedge",
    "surfinterior",
    "surfpart",
    "surfplane",
    "surfreorient",
    "surfseeds",
    "surfvolume",
    "uniqedges",
    "uniqfaces",
    "v2m",
    "v2s",
    "varargin2struct",
    "vol2mesh",
    "vol2restrictedtri",
    "vol2surf",
    "volface",
    "linextriangle",
    "getplanefrom3pt",
    "polylinelen",
    "closestnode",
    "polylineinterp",
    "polylinesimplify",
    "affinemap",
    "meshinterp",
    "meshremap",
    "proj2mesh",
    "dist2surf",
    "regpt2surf",
    "ismember_rows",
    "volgrow",
    "volshrink",
    "volopen",
    "volclose",
    "fillholes3d",
    "thickenbinvol",
    "thinbinvol",
    "brain2mesh",
    "brain1020",
    "intriangulation",
    "label2tpm",
    "tpm2label",
    "ray2surf",
    "nargout",
    "loadjson",
    "savejson",
    "loadjd",
    "savejd",
    "loadbj",
    "savebj",
    "jsonpath",
    "jdataencode",
    "jdatadecode",
    "zlibencode",
    "zlibdecode",
    "gzipencode",
    "gzipdecode",
    "lzmaencode",
    "lzmadecode",
    "lz4encode",
    "lz4decode",
    "base64encode",
    "base64decode",
    "nii2jnii",
    "jnii2nii",
    "loadnifti",
    "loadjnifti",
    "savenifti",
    "savejnifti",
    "nifticreate",
    "jnifticreate",
    "memmapstream",
    "niiheader2jnii",
    "niicodemap",
    "niiformat",
    "jdlink",
    "jsoncache",
    "loadurl",
]
__license__ = """GNU General Public License v3 and later"""
